from numpy.testing import assert_almost_equal, assert_array_equal, \
    assert_array_almost_equal
from ctapipe.calib.camera.r1 import (
    CameraR1CalibratorFactory,
    HESSIOR1Calibrator,
    TargetIOR1Calibrator,
    NullR1Calibrator
)
from ctapipe.io.hessioeventsource import HESSIOEventSource
from ctapipe.io.targetioeventsource import TargetIOEventSource
from ctapipe.io.eventsource import EventSource
from ctapipe.utils import get_dataset_path
from copy import deepcopy
import pytest


def test_hessio_r1_calibrator(test_event):
    telid = 11
    event = deepcopy(test_event)
    calibrator = HESSIOR1Calibrator()
    calibrator.calibrate(event)
    r1 = event.r1.tel[telid].waveform
    assert_almost_equal(r1[0, 0, 0], -0.091, 3)


def test_null_r1_calibrator(test_event):
    telid = 11
    event = deepcopy(test_event)
    calibrator = NullR1Calibrator()
    calibrator.calibrate(event)
    r0 = event.r0.tel[telid].waveform
    r1 = event.r1.tel[telid].waveform
    assert_array_equal(r0, r1)


def test_targetio_calibrator():
    pytest.importorskip("target_calib")
    url_r0 = get_dataset_path("targetmodule_r0.tio")
    url_r1 = get_dataset_path("targetmodule_r1.tio")
    pedpath = get_dataset_path("targetmodule_ped.tcal")

    source_r0 = TargetIOEventSource(input_url=url_r0)
    source_r1 = TargetIOEventSource(input_url=url_r1)

    r1c = CameraR1CalibratorFactory.produce(eventsource=source_r0)

    event_r0 = source_r0._get_event_by_index(0)
    event_r1 = source_r1._get_event_by_index(0)

    r1c.calibrate(event_r0)
    assert_array_equal(event_r0.r0.tel[0].waveform,
                       event_r0.r1.tel[0].waveform)

    r1c = CameraR1CalibratorFactory.produce(
        eventsource=source_r0,
        pedestal_path=pedpath
    )
    r1c.calibrate(event_r0)
    assert_array_almost_equal(event_r0.r1.tel[0].waveform,
                              event_r1.r1.tel[0].waveform, 1)


def test_targetio_calibrator_wrong_file(test_event):
    pytest.importorskip("target_calib")
    r1c = TargetIOR1Calibrator()
    with pytest.raises(ValueError):
        r1c.calibrate(test_event)


def test_check_r0_exists(test_event):
    telid = 11
    event = deepcopy(test_event)
    calibrator = HESSIOR1Calibrator()
    assert(calibrator.check_r0_exists(event, telid) is True)
    event.r0.tel[telid].waveform = None
    assert(calibrator.check_r0_exists(event, telid) is False)


def test_factory_from_product():
    calibrator = CameraR1CalibratorFactory.produce(
        product="NullR1Calibrator"
    )
    assert isinstance(calibrator, NullR1Calibrator)
    calibrator = CameraR1CalibratorFactory.produce(
        product="HESSIOR1Calibrator"
    )
    assert isinstance(calibrator, HESSIOR1Calibrator)


def test_factory_default():
    calibrator = CameraR1CalibratorFactory.produce()
    assert isinstance(calibrator, NullR1Calibrator)


def test_factory_from_eventsource():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    eventsource = HESSIOEventSource(input_url=dataset)
    calibrator = CameraR1CalibratorFactory.produce(eventsource=eventsource)
    assert isinstance(calibrator, HESSIOR1Calibrator)


def test_factory_from_eventsource_override():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    eventsource = HESSIOEventSource(input_url=dataset)
    calibrator = CameraR1CalibratorFactory.produce(
        eventsource=eventsource,
        product="NullR1Calibrator"
    )
    assert isinstance(calibrator, NullR1Calibrator)


class UnknownEventSource(EventSource):
    """
    Simple working EventSource
    """

    def _generator(self):
        return range(len(self.input_url))

    @staticmethod
    def is_compatible(file_path):
        return False


def test_factory_from_unknown_eventsource():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    eventsource = UnknownEventSource(input_url=dataset)
    calibrator = CameraR1CalibratorFactory.produce(eventsource=eventsource)
    assert isinstance(calibrator, NullR1Calibrator)



"""
Classes and functions related to telescope Optics
"""

import logging
from ..utils import get_table_dataset
import numpy as np
import astropy.units as u

logger = logging.getLogger(__name__)

_FOCLEN_TO_TEL_INFO = {
    # foclen: tel_type, tel_subtype, mirror_type
    28.0: ('LST', '', 'DC'),
    16.0: ('MST', '', 'DC'),
    2.28: ('SST', 'GCT', 'SC'),
    2.15: ('SST', 'ASTRI', 'SC'),
    5.6: ('SST', '1M', 'SC'),
    5.58: ('MST', 'SCT', 'SC'),
    5.59: ('MST', 'SCT', 'SC')

}


class OpticsDescription:
    """
    Describes the optics of a Cherenkov Telescope mirror

    The string representation of an `OpticsDescription` will be a combination
    of the telescope-type and sub-type as follows: "type-subtype". You can
    also get each individually.

    The `OpticsDescription.guess()` constructor can be used to fill in info
    from metadata, e.g. for Monte-Carlo files.

    Parameters
    ----------
    mirror_type: str
        'SC' or 'DC'
    tel_type: str
        'SST', 'MST','LST'
    tel_subtype: str
        subtype of telescope, e.g. '1M' or 'ASTRI'
    equivalent_focal_length: Quantity(float)
        effective focal-length of telescope, independent of which type of
        optics (as in the Monte-Carlo)
    mirror_area: float
        total reflective surface area of the optical system (in m^2)
    num_mirror_tiles: int
        number of mirror facets

    Raises
    ------
    ValueError:
        if tel_type or mirror_type are not one of the accepted values
    TypeError, astropy.units.UnitsError:
        if the units of one of the inputs are missing or incompatible
    """

    def __init__(self, mirror_type, tel_type,
                 tel_subtype, equivalent_focal_length,
                 mirror_area=None, num_mirror_tiles=None):

        if tel_type not in ['LST', 'MST', 'SST']:
            raise ValueError("Unknown tel_type %s", tel_type)

        if mirror_type not in ['SC', 'DC']:
            raise ValueError("Unknown mirror_type: %s", mirror_type)

        self.mirror_type = mirror_type
        self.tel_type = tel_type
        self.tel_subtype = tel_subtype
        self.equivalent_focal_length = equivalent_focal_length.to(u.m)
        self.mirror_area = mirror_area

        self.num_mirror_tiles = num_mirror_tiles

    @classmethod
    @u.quantity_input
    def guess(cls, equivalent_focal_length: u.m):
        """
        Construct an OpticsDescription by guessing from metadata (e.g. when
        using a simulation where the exact type is not known)

        Parameters
        ----------
        equivalent_focal_length: Quantity('m')
            effective optical focal-length in meters

        Raises
        ------
        KeyError:
            on unknown focal length
        """

        tel_type, tel_subtype, mir_type = \
            telescope_info_from_metadata(equivalent_focal_length)

        return cls(mirror_type=mir_type,
                   tel_type=tel_type,
                   tel_subtype=tel_subtype,
                   equivalent_focal_length=equivalent_focal_length)

    @classmethod
    def from_name(cls, name, optics_table='optics'):
        """
        Construct an OpticsDescription from the name. This is loaded from
        `optics.fits.gz`, which should be in `ctapipe_resources` or in a
        directory listed in `CTAPIPE_SVC_PATH`.

        Parameters
        ----------
        name: str
            string representation of optics (MST, LST, SST-1M, SST-ASTRI,...)
        optics_table: str
            base filename of optics table if not 'optics.*'


        Returns
        -------
        OpticsDescription

        """
        table = get_table_dataset(optics_table,
                                  role='dl0.tel.svc.optics')
        mask = table['tel_description'] == name

        if 'equivalent_focal_length' in table.colnames:
            flen = table['equivalent_focal_length'][mask].quantity[0]
        else:
            flen = table['effective_focal_length'][mask].quantity[0]
            logger.warning("Optics table format out of date: "
                           "'effective_focal_length' "
                           "should be 'equivalent_focal_length'")

        optics = cls(
            mirror_type=table['mirror_type'][mask][0],
            tel_type=table['tel_type'][mask][0],
            tel_subtype=table['tel_subtype'][mask][0],
            equivalent_focal_length=flen,
            mirror_area=table['mirror_area'][mask].quantity[0],
            num_mirror_tiles=table['num_mirror_tiles'][mask][0],
        )
        return optics

    @classmethod
    def get_known_optics_names(cls, optics_table='optics'):
        table = get_table_dataset(optics_table, 'get_known_optics')
        return np.array(table['tel_description'])

    @property
    def identifier(self):
        """ returns a tuple of (tel_type, tel_subtype).  Use str(optics) to
        get a text-based identifier."""
        return self.tel_type, self.tel_subtype

    def info(self, printer=print):
        printer('OpticsDescription: "{}"'.format(self))
        printer('    - mirror_type: {}'.format(self.mirror_type))
        printer('    - num_mirror_tiles: {}'.format(self.num_mirror_tiles))
        printer('    - mirror_area: {}'.format(self.mirror_area))

    def __repr__(self):
        return "{}(tel_type='{}', tel_subtype='{}')".format(
            str(self.__class__.__name__), self.tel_type, self.tel_subtype)

    def __str__(self):
        if self.tel_subtype != '':
            return "{}-{}".format(self.tel_type, self.tel_subtype)
        else:
            return self.tel_type


def telescope_info_from_metadata(focal_length):
    """
    helper func to return telescope and mirror info based on metadata

    Parameters
    ----------
    focal_length: float
        effective focal length

    Returns
    -------
    str,str,str:
        tel_type ('LST', 'MST' or 'SST'),
        tel_subtype (model),
        mirror_type ('SC' or 'DC')

    Raises:
    -------
    KeyError:
       if unable to find optics type
    """
    return _FOCLEN_TO_TEL_INFO[round(focal_length.to('m').value, 2)]



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
EventSource for NectarCam protobuf-fits.fz-files.

Needs protozfits v0.44.5 from github.com/cta-sst-1m/protozfitsreader
"""

import numpy as np
from .eventsource import EventSource
from .containers import NectarCAMDataContainer

__all__ = ['NectarCAMEventSource']


class NectarCAMEventSource(EventSource):

    def __init__(self, config=None, tool=None, **kwargs):
        super().__init__(config=config, tool=tool, **kwargs)
        from protozfits import SimpleFile
        self.file = SimpleFile(self.input_url)
        self.header = next(self.file.RunHeader)


    def _generator(self):

        self._pixel_sort_ids = None

        for count, event in enumerate(self.file.Events):
            data = NectarCAMDataContainer()
            data.count = count
            # fill specific NectarCAM data
            data.nectarcam.fill_from_zfile_event(event, self.header.numTraces)
            # fill general R0 data
            self.fill_R0Container_from_zfile_event(data.r0, event)
            yield data


    @staticmethod
    def is_compatible(file_path):
        from astropy.io import fits
        try:
            # The file contains two tables:
            #  1: RunHeader
            #  2: Events <--- this is what we need to look at
            h = fits.open(file_path)[2].header
            ttypes = [
                h[x] for x in h.keys() if 'TTYPE' in x
            ]
        except OSError:
            # not even a fits file
            return False

        except IndexError:
            # A fits file of a different format
            return False

        is_protobuf_zfits_file = (
            (h['XTENSION'] == 'BINTABLE') and
            (h['EXTNAME'] == 'Events') and
            (h['ZTABLE'] is True) and
            (h['ORIGIN'] == 'CTA') and
            (h['PBFHEAD'] == 'DataModel.CameraEvent')
        )

        is_nectarcam_file = 'hiGain_integrals_gains' in ttypes
        return is_protobuf_zfits_file & is_nectarcam_file


    def fill_R0CameraContainer_from_zfile_event(self, container, event):
        container.trigger_time = (
            event.local_time_sec * 1E9 + event.local_time_nanosec)
        container.trigger_type = event.event_type

        container.waveform = np.array([
            (
                event.hiGain.waveforms.samples
            ).reshape(-1, self.header.numTraces),
            (
                event.loGain.waveforms.samples
            ).reshape(-1, self.header.numTraces)
        ])

        container.num_samples = container.waveform.shape[1]

    def fill_R0Container_from_zfile_event(self, container, event):
        container.obs_id = -1
        container.event_id = event.eventNumber

        container.tels_with_data = [self.header.telescopeID, ]
        r0_cam_container = container.tel[self.header.telescopeID]
        self.fill_R0CameraContainer_from_zfile_event(
            r0_cam_container,
            event
        )



"""
Conveniance module to handle the calling of functions within
neighbour_sum_c.cc through ctypes.
"""

__all__ = []

import numpy as np
import ctypes
from numpy.ctypeslib import ndpointer
import os

try:
    lib = np.ctypeslib.load_library("neighbour_sum_c", os.path.dirname(__file__))
    get_sum_array = lib.get_sum_array
    get_sum_array.restype = None
    get_sum_array.argtypes = [ndpointer(ctypes.c_float, flags="C_CONTIGUOUS"),
                              ndpointer(ctypes.c_float, flags="C_CONTIGUOUS"),
                              ctypes.c_size_t, ctypes.c_size_t, ctypes.c_size_t,
                              ndpointer(ctypes.c_uint16, flags="C_CONTIGUOUS"),
                              ctypes.c_size_t,
                              ctypes.c_int]
except OSError:
    pass



from ctapipe.instrument.optics import OpticsDescription
from astropy import units as u
import pytest


def test_guess_optics():
    od = OpticsDescription.guess(28.0 * u.m)
    od.info()

    assert od.tel_type == 'LST'
    assert od.tel_subtype == ''
    assert od.mirror_type == 'DC'

    with pytest.raises(KeyError):
        OpticsDescription.guess(0 * u.m)  # unknown tel


def test_construct_optics():
    with pytest.raises(ValueError):
        OpticsDescription(mirror_type="DC",
                          tel_type="bad",  # bad value
                          tel_subtype="1M",
                          equivalent_focal_length=10 * u.m)

    with pytest.raises(ValueError):
        OpticsDescription(mirror_type="bad",  # bad value
                          tel_type="MST",
                          tel_subtype="1M",
                          equivalent_focal_length=10 * u.m)

    with pytest.raises(u.UnitsError):
        OpticsDescription.guess(28.0 * u.kg)  # bad unit

    with pytest.raises(TypeError):
        OpticsDescription.guess(28.0)  # not a unit quantity



from copy import deepcopy

from numpy.testing import assert_allclose

from ctapipe.calib.camera.dl0 import CameraDL0Reducer
from ctapipe.calib.camera.dl1 import integration_correction, \
    CameraDL1Calibrator
from ctapipe.calib.camera.r1 import HESSIOR1Calibrator


def previous_calibration(event):
    r1 = HESSIOR1Calibrator()
    r1.calibrate(event)
    dl0 = CameraDL0Reducer()
    dl0.reduce(event)


def test_integration_correction(test_event):
    event = deepcopy(test_event)
    telid = 11
    width = 7
    shift = 3
    shape = event.mc.tel[telid].reference_pulse_shape
    n_chan = shape.shape[0]
    step = event.mc.tel[telid].meta['refstep']
    time_slice = event.mc.tel[telid].time_slice
    correction = integration_correction(n_chan, shape, step,
                                        time_slice, width, shift)
    assert_allclose(correction[0], 1.077, 1e-3)


def test_integration_correction_no_ref_pulse(test_event):
    event = deepcopy(test_event)
    previous_calibration(event)
    telid = list(event.dl0.tel.keys())[0]
    delattr(event, 'mc')
    calibrator = CameraDL1Calibrator()
    correction = calibrator.get_correction(event, telid)
    assert correction[0] == 1


def test_camera_dl1_calibrator(test_event):
    event = deepcopy(test_event)
    previous_calibration(event)
    telid = 11

    calibrator = CameraDL1Calibrator()

    correction = calibrator.get_correction(event, telid)
    assert_allclose(correction[0], 1.077, 1e-3)

    calibrator.calibrate(event)
    image = event.dl1.tel[telid].image
    assert_allclose(image[0, 0], -2.216, 1e-3)


def test_check_dl0_exists(test_event):
    event = deepcopy(test_event)
    telid = 11
    previous_calibration(event)
    calibrator = CameraDL1Calibrator()
    assert(calibrator.check_dl0_exists(event, telid) is True)
    event.dl0.tel[telid].waveform = None
    assert(calibrator.check_dl0_exists(event, telid) is False)



from copy import deepcopy

from numpy.testing import assert_almost_equal

from ctapipe.calib.camera.dl0 import CameraDL0Reducer
from ctapipe.calib.camera.r1 import HESSIOR1Calibrator


def previous_calibration(event):
    r1 = HESSIOR1Calibrator()
    r1.calibrate(event)


def test_camera_dl0_reducer(test_event):
    event = deepcopy(test_event)
    previous_calibration(event)
    telid = 11

    reducer = CameraDL0Reducer()
    reducer.reduce(event)
    waveforms = event.dl0.tel[telid].waveform
    assert_almost_equal(waveforms[0, 0, 0], -0.091, 3)


def test_check_r1_exists(test_event):
    event = deepcopy(test_event)
    telid = 11
    previous_calibration(event)
    reducer = CameraDL0Reducer()
    assert(reducer.check_r1_exists(event, telid) is True)
    event.r1.tel[telid].waveform = None
    assert(reducer.check_r1_exists(event, telid) is False)



from ctapipe.image.timing_parameters import timing_parameters
import numpy as np
import astropy.units as u
from numpy.testing import assert_allclose


def test_grad_fit():
    """
    Simple test that gradient fitting gives expected answers for perfect
    gradient
    """
    grad = 2
    intercept = 1

    timing = timing_parameters(
        pix_x=np.zeros(4) * u.deg,
        pix_y=np.arange(4) * u.deg,
        image=np.ones(4),
        peak_time=intercept * u.ns + grad * np.arange(4) * u.ns,
        rotation_angle=0 * u.deg
    )

    # Test we get the values we put in back out again
    assert_allclose(timing.gradient, grad * u.ns / u.deg)
    assert_allclose(timing.intercept, intercept * u.deg)

    # Then try a different rotation angle
    rot_angle = 20 * u.deg
    timing_rot20 = timing_parameters(
        pix_x=np.zeros(4) * u.deg,
        pix_y=np.arange(4) * u.deg,
        image=np.ones(4),
        peak_time=intercept * u.ns +
        grad * np.arange(4) * u.ns,
        rotation_angle=rot_angle
    )
    # Test the output again makes sense
    assert_allclose(timing_rot20.gradient, timing.gradient / np.cos(rot_angle))
    assert_allclose(timing_rot20.intercept, timing.intercept)



from ctapipe.io.hessio import hessio_event_source
from ctapipe.utils import get_dataset_path
from ctapipe.io.hessioeventsource import HESSIOEventSource


def test_hessio_event_source():
    filename = get_dataset_path("gamma_test.simtel.gz")
    source = hessio_event_source(filename)
    event = next(source)
    tels = event.dl0.tels_with_data
    print(tels)
    assert tels == {38, 47}




"""
Set of diagnostic plots relating to muons 
For generic use with all muon algorithms
"""

import matplotlib.mlab as mlab
import matplotlib.pyplot as plt
import numpy as np
from astropy import units as u
from astropy.table import Table
from matplotlib import colors
from scipy.stats import norm

from ctapipe.coordinates import (CameraFrame, NominalFrame, HorizonFrame)
from ctapipe.image.cleaning import tailcuts_clean
from ctapipe.plotting.camera import CameraPlotter
from ctapipe.utils.fitshistogram import Histogram

import logging

logger = logging.getLogger(__name__)


def plot_muon_efficiency(outputpath):
    """
    Plot the muon efficiencies
    """
    fig, ax = plt.subplots(1, 1, figsize=(10, 10))
    figip, axip = plt.subplots(1, 1, figsize=(10, 10))
    figrw, axrw = plt.subplots(1, 1, figsize=(10, 10))

    nbins = 16
    t = Table.read(str(outputpath) + '_muontable.fits')
    logger.info('Reading muon efficiency from table "%s"', outputpath)

    if len(t['MuonEff']) < 1:
        logger.warning("No muon events to plot")
        return
    else:
        logger.info("Found %d muon events", len(t['MuonEff']))

    (mu, sigma) = norm.fit(t['MuonEff'])

    logger.debug('Gaussian fit with mu=%f, sigma=%f', mu, sigma)

    conteff = ax.hist(t['MuonEff'], nbins)
    ax.set_xlim(0.2 * min(t['MuonEff']), 1.2 * max(t['MuonEff']))

    xtest = np.linspace(min(t['MuonEff']), max(t['MuonEff']), nbins)
    yg = mlab.normpdf(xtest, mu, sigma)
    logger.debug('mu=%f sigma=%f yg=%f', mu, sigma, yg)
    ax.plot(xtest, yg, 'r', linewidth=2)

    ax.set_ylim(0., 1.2 * max(conteff[0]))
    ax.set_xlabel('Muon Efficiency')
    plt.draw()

    contimp = axip.hist(t['ImpactP'], nbins)
    axip.set_xlim(0.2 * min(t['ImpactP']), 1.2 * max(t['ImpactP']))
    axip.set_ylim(0., 1.2 * max(contimp[0]))
    axip.set_xlabel('Impact Parameter (m)')

    plt.draw()

    heffimp = Histogram(nbins=[16, 16],
                        ranges=[(min(t['MuonEff']), max(t['MuonEff'])),
                                (min(t['ImpactP']), max(t[
                                    'ImpactP']))])  #
    #  ,axisNames=["MuonEfficiency","ImpactParameter"])

    heffimp.fill([t['MuonEff'], t['ImpactP']])
    heffimp.draw_2d()

    contrw = axrw.hist(t['RingWidth'], nbins)
    axrw.set_xlim(0.2 * min(t['RingWidth']), 1.2 * max(t['RingWidth']))
    axrw.set_ylim(0., 1.2 * max(contrw[0]))
    axrw.set_xlabel('Ring Width ($^\circ$)')

    plt.draw()

    if outputpath is not None:
        logger.info("saving figure to '%s'", outputpath)
        fig.savefig(str(outputpath) + '_MuonEff.png')
        figip.savefig(str(outputpath) + '_ImpactParameter.png')
        figrw.savefig(str(outputpath) + '_RingWidth.png')
    else:
        logger.info("Not saving figure, no outputpath")
        plt.show()


def plot_muon_event(event, muonparams):
    if muonparams['MuonRingParams'] is not None:

        # Plot the muon event and overlay muon parameters
        fig = plt.figure(figsize=(16, 7))

        colorbar = None
        colorbar2 = None

        subarray = event.inst.subarray

        # for tel_id in event.dl0.tels_with_data:
        for tel_id in muonparams['TelIds']:
            idx = muonparams['TelIds'].index(tel_id)

            if not muonparams['MuonRingParams'][idx]:
                continue

            # otherwise...
            npads = 2
            # Only create two pads if there is timing information extracted
            # from the calibration
            ax1 = fig.add_subplot(1, npads, 1)
            plotter = CameraPlotter(event)
            image = event.dl1.tel[tel_id].image[0]
            geom = event.inst.subarray.tel[tel_id].camera

            tailcuts = (5., 7.)
            # Try a higher threshold for
            if geom.cam_id == 'FlashCam':
                tailcuts = (10., 12.)

            clean_mask = tailcuts_clean(geom, image,
                                        picture_thresh=tailcuts[0],
                                        boundary_thresh=tailcuts[1])

            signals = image * clean_mask

            rotr_angle = geom.pix_rotation
# The following two lines have been commented out to avoid a rotation error.
#            if geom.cam_id == 'LSTCam' or geom.cam_id == 'NectarCam':

#                rotr_angle = 0. * u.deg

            # Convert to camera frame (centre & radius)
            altaz = HorizonFrame(alt=event.mc.alt, az=event.mc.az)

            ring_nominal = NominalFrame(
                x=muonparams['MuonRingParams'][idx].ring_center_x,
                y=muonparams['MuonRingParams'][idx].ring_center_y,
                array_direction=altaz,
                pointing_direction=altaz)

            flen = subarray.tel[tel_id].optics.equivalent_focal_length
            ring_camcoord = ring_nominal.transform_to(CameraFrame(
                pointing_direction=altaz,
                focal_length=flen,
                rotation=rotr_angle))

            centroid = (ring_camcoord.x.value, ring_camcoord.y.value)

            ringrad_camcoord = muonparams['MuonRingParams'][idx].ring_radius.to(
                u.rad) \
                * flen * 2.  # But not FC?

            px = subarray.tel[tel_id].camera.pix_x
            py = subarray.tel[tel_id].camera.pix_y
            camera_coord = CameraFrame(x=px, y=py,
                                       focal_length=flen,
                                       rotation=geom.pix_rotation)

            nom_coord = camera_coord.transform_to(
                NominalFrame(array_direction=altaz,
                             pointing_direction=altaz)
            )

            px = nom_coord.x.to(u.deg)
            py = nom_coord.y.to(u.deg)
            dist = np.sqrt(np.power(px - muonparams['MuonRingParams'][idx].ring_center_x,
                                    2) + np.power(py - muonparams['MuonRingParams'][idx].
                                                  ring_center_y, 2))
            ring_dist = np.abs(dist - muonparams['MuonRingParams'][idx].ring_radius)
            pix_rmask = ring_dist < muonparams['MuonRingParams'][idx].ring_radius * 0.4

            if muonparams['MuonIntensityParams'][idx] is not None:
                signals *= muonparams['MuonIntensityParams'][idx].mask
            elif muonparams['MuonIntensityParams'][idx] is None:
                signals *= pix_rmask

            camera1 = plotter.draw_camera(tel_id, signals, ax1)

            cmaxmin = (max(signals) - min(signals))
            cmin = min(signals)
            if not cmin:
                cmin = 1.
            if not cmaxmin:
                cmaxmin = 1.

            cmap_charge = colors.LinearSegmentedColormap.from_list(
                'cmap_c', [(0 / cmaxmin, 'darkblue'),
                           (np.abs(cmin) / cmaxmin, 'black'),
                           (2.0 * np.abs(cmin) / cmaxmin, 'blue'),
                           (2.5 * np.abs(cmin) / cmaxmin, 'green'),
                           (1, 'yellow')]
            )
            camera1.pixels.set_cmap(cmap_charge)
            if not colorbar:
                camera1.add_colorbar(ax=ax1, label=" [photo-electrons]")
                colorbar = camera1.colorbar
            else:
                camera1.colorbar = colorbar
            camera1.update(True)

            camera1.add_ellipse(centroid, ringrad_camcoord.value,
                                ringrad_camcoord.value, 0., 0., color="red")

            if muonparams['MuonIntensityParams'][idx] is not None:

                ringwidthfrac = muonparams['MuonIntensityParams'][idx].ring_width / \
                    muonparams['MuonRingParams'][idx].ring_radius
                ringrad_inner = ringrad_camcoord * (1. - ringwidthfrac)
                ringrad_outer = ringrad_camcoord * (1. + ringwidthfrac)
                camera1.add_ellipse(centroid, ringrad_inner.value,
                                    ringrad_inner.value, 0., 0.,
                                    color="magenta")
                camera1.add_ellipse(centroid, ringrad_outer.value,
                                    ringrad_outer.value, 0., 0.,
                                    color="magenta")
                npads = 2
                ax2 = fig.add_subplot(1, npads, npads)
                pred = muonparams['MuonIntensityParams'][idx].prediction

                if len(pred) != np.sum(
                        muonparams['MuonIntensityParams'][idx].mask):
                    logger.warning("Lengths do not match...len(pred)=%s len("
                                   "mask)=", len(pred),
                                   np.sum(muonparams['MuonIntensityParams'][idx].mask))

                # Numpy broadcasting - fill in the shape
                plotpred = np.zeros(image.shape)
                truelocs = np.where(muonparams['MuonIntensityParams'][idx].mask == True)
                plotpred[truelocs] = pred

                camera2 = plotter.draw_camera(tel_id, plotpred, ax2)

                if np.isnan(max(plotpred)) or np.isnan(min(plotpred)):
                    logger.debug("nan prediction, skipping...")
                    continue

                c2maxmin = (max(plotpred) - min(plotpred))
                if not c2maxmin:
                    c2maxmin = 1.

                c2map_charge = colors.LinearSegmentedColormap.from_list(
                    'c2map_c', [(0 / c2maxmin, 'darkblue'),
                                (np.abs(min(plotpred)) / c2maxmin, 'black'),
                                (
                                2.0 * np.abs(min(plotpred)) / c2maxmin, 'blue'),
                                (2.5 * np.abs(min(plotpred)) / c2maxmin,
                                 'green'),
                                (1, 'yellow')]
                )
                camera2.pixels.set_cmap(c2map_charge)
                if not colorbar2:
                    camera2.add_colorbar(ax=ax2, label=" [photo-electrons]")
                    colorbar2 = camera2.colorbar
                else:
                    camera2.colorbar = colorbar2
                camera2.update(True)
                plt.pause(1.)  # make shorter

            # plt.pause(0.1)
            # if pp is not None:
            #    pp.savefig(fig)
            # fig.savefig(str(args.output_path) + "_" +
            #            str(event.dl0.event_id) + '.png')


            plt.close()



from ctapipe.core.factory import Factory
from ctapipe.core.component import Component
from traitlets import Int, TraitError
import pytest
from traitlets.config.loader import Config
import warnings


class ExampleComponentParent(Component):
    value = Int(123, help="").tag(config=True)

    def __init__(self, config, parent, **kwargs):
        super().__init__(config=config, parent=parent, **kwargs)


class ExampleComponent1(ExampleComponentParent):
    value = Int(123111, help="").tag(config=True)


class ExampleComponent2(ExampleComponentParent):
    value = Int(123222, help="").tag(config=True)


class ExampleComponent3(ExampleComponentParent):
    value = Int(123333, help="").tag(config=True)

    def __init__(self, config, parent, extra=0, **kwargs):
        super().__init__(config=config, parent=parent, **kwargs)
        self.extra = extra


class ExampleComponent4(ExampleComponentParent):
    value = Int(123444, help="").tag(config=True)
    extra = Int(5, help="").tag(config=True)


class ExampleFactory(Factory):
    base = ExampleComponentParent
    default = 'ExampleComponent1'


class IncorrectExampleFactory(Factory):
    base = ExampleComponentParent
    default = 'ExampleComponent1'

    def _get_product_name(self):
        return "NonExistantClass"


def test_factory():
    obj = ExampleFactory.produce(
        config=None, tool=None,
        product='ExampleComponent2',
        value=111
    )
    assert(obj.__class__.__name__ == 'ExampleComponent2')
    assert(obj.value == 111)


def test_factory_subclass_detection():
    subclasses = [
        ExampleComponent1,
        ExampleComponent2,
        ExampleComponent3,
        ExampleComponent4
    ]
    subclass_names = [
        "ExampleComponent1",
        "ExampleComponent2",
        "ExampleComponent3",
        "ExampleComponent4"
    ]
    factory_subclasses_str = [str(i) for i in ExampleFactory.subclasses]
    subclasses_str = [str(i) for i in subclasses]
    assert sorted(factory_subclasses_str) == sorted(subclasses_str)
    assert sorted(ExampleFactory.subclass_names) == sorted(subclass_names)


def test_factory_automatic_traits():
    traits = sorted(list(ExampleFactory.class_own_traits().keys()))
    assert traits == sorted(['extra', 'product', 'value'])


def test_factory_traits_compatible_help():
    msg = [
        "Compatible Components:",
        "ExampleComponent1",
        "ExampleComponent2",
        "ExampleComponent3",
        "ExampleComponent4"
        ]
    for m in msg:
        assert m in ExampleFactory.class_own_traits()['value'].help


def test_factory_produce():
    obj = ExampleFactory.produce(config=None, tool=None,
                                 product='ExampleComponent2',
                                 value=111)
    assert (obj.__class__.__name__ == 'ExampleComponent2')
    assert (obj.value == 111)


def test_false_product_name():
    with pytest.raises(KeyError):
        IncorrectExampleFactory.produce(
            config=None, tool=None,
            product='ExampleComponent2',
            value=111
        )


def test_expected_args():
    kwargs = dict(
        product='ExampleComponent2',
        value=111,
        extra=4,
        nonexistant=5
    )
    with pytest.raises(TraitError):
        obj = ExampleFactory.produce(config=None, tool=None, **kwargs)

    kwargs.pop('nonexistant')
    obj = ExampleFactory.produce(config=None, tool=None, **kwargs)

    with pytest.raises(AttributeError):
        assert obj.extra == 4
    with pytest.raises(AttributeError):
        assert obj.nonexistant == 5

    kwargs['product'] = 'ExampleComponent3'
    obj = ExampleFactory.produce(config=None, tool=None, **kwargs)
    assert obj.extra == 4
    with pytest.raises(AttributeError):
        assert obj.nonexistant == 5

    kwargs['product'] = 'ExampleComponent4'
    obj = ExampleFactory.produce(config=None, tool=None, **kwargs)
    assert obj.extra == 4
    with pytest.raises(AttributeError):
        assert obj.nonexistant == 5


def test_expected_config():
    config = Config()
    config['ExampleFactory'] = Config()
    config['ExampleFactory']['product'] = 'ExampleComponent2'
    config['ExampleFactory']['value'] = 111
    config['ExampleFactory']['extra'] = 4
    obj = ExampleFactory.produce(config=config, tool=None)
    assert obj.value == 111
    with pytest.raises(AttributeError):
        assert obj.extra == 4

    config = Config()
    config['ExampleFactory'] = Config()
    config['ExampleFactory']['product'] = 'ExampleComponent2'
    config['ExampleComponent2'] = Config()
    config['ExampleComponent2']['value'] = 111
    config['ExampleComponent2']['extra'] = 4
    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=UserWarning)
        obj = ExampleFactory.produce(config=config, tool=None)
    assert obj.value == 111
    with pytest.raises(AttributeError):
        assert obj.extra == 4

    config = Config()
    config['ExampleFactory'] = Config()
    config['ExampleFactory']['product'] = 'ExampleComponent4'
    config['ExampleFactory']['value'] = 111
    config['ExampleFactory']['extra'] = 4
    obj = ExampleFactory.produce(config=config, tool=None)
    assert obj.value == 111
    assert obj.extra == 4

    config['ExampleFactory'] = Config()
    config['ExampleFactory']['product'] = 'ExampleComponent4'
    config['ExampleComponent4'] = Config()
    config['ExampleComponent4']['value'] = 111
    config['ExampleComponent4']['extra'] = 4
    obj = ExampleFactory.produce(config=config, tool=None)
    assert obj.value == 111
    assert obj.extra == 4


def test_default_passing():
    old_default = ExampleFactory.value.default_value
    ExampleFactory.value.default_value = 3
    obj = ExampleFactory.produce(config=None, tool=None,
                                 product='ExampleComponent2')
    assert (obj.value == 3)
    obj = ExampleFactory.produce(config=None, tool=None,
                                 product='ExampleComponent4')
    assert (obj.value == 3)
    ExampleFactory.value.default_value = old_default
    obj = ExampleFactory.produce(config=None, tool=None)
    assert (obj.value == old_default)



import logging

import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier

from .regressor_classifier_base import RegressorClassifierBase

__all__ = ['proba_drifting','EventClassifier']


def proba_drifting(x):
    """
    gives more weight to outliers -- i.e. close to 0 and 1
    the polynomial was constructed with the following constraints:
    • f(0) = 0
    • f(0.5) = 0.5
    • f(1) = 1
    • f'(0) = 0
    • f'(0.5) = 1
    • f'(1) = 0
    """
    return 10 * x ** 3 - 15 * x ** 4 + 6 * x ** 5


class EventClassifier(RegressorClassifierBase):

    def __init__(self, classifier=RandomForestClassifier,
                 cam_id_list=("cam"), **kwargs):
        super().__init__(model=classifier, cam_id_list=cam_id_list, **kwargs)

    def predict_proba_by_event(self, X):
        predict_proba = []
        for evt in X:
            tel_probas = None
            tel_weights = []
            for cam_id, tels in evt.items():
                these_probas = self.model_dict[cam_id].predict_proba(tels)
                if tel_probas is not None:
                    tel_probas = np.append(these_probas, tel_probas, axis=0)
                else:
                    tel_probas = these_probas
                try:
                    # if a `namedtuple` is provided, we can weight the
                    # different images using some of the provided features
                    tel_weights += [t.sum_signal_cam / t.impact_dist for t in
                                    tels]
                except AttributeError:
                    # otherwise give every image the same weight
                    tel_weights += [1] * len(tels)

            predict_proba.append(
                np.average(proba_drifting(tel_probas),
                           weights=tel_weights,
                           axis=0)
            )

        return np.array(predict_proba)

    def predict_by_event(self, X):
        proba = self.predict_proba_by_event(X)
        predictions = self.classes_[np.argmax(proba, axis=1)]
        return predictions

    def compute_Qfactor(self, proba, labels: int, nbins):
        """
        Compute Q-factor for each gammaness (bin edges are 0 - 1)

        Parameters
        ----------

        proba: predicted probabilities to be a gamma!
        labels: true labels
        nbins: number of bins for gammaness

        Returns
        -------

        Q-factor array
        """
        bins = np.linspace(0, 1, nbins)

        # assuming labels are 0 for protons, 1 for gammas
        # np.nonzero function return indexes
        gammas_idx = np.nonzero(proba * labels)
        gammas = proba[gammas_idx]

        hadrons_idx = np.nonzero(proba * np.logical_not(labels))
        hadrons = proba[hadrons_idx]

        # tot number of gammas
        Ng = len(gammas)
        # of protons
        Nh = len(hadrons)

        # binning and cumsum for gammas
        gbins = pd.cut(gammas, bins)
        gcount = gbins.value_counts()
        # reverse array
        gcount[:] = gcount.values[::-1].copy()
        g_cumsum = np.cumsum(gcount.values)
        eps_g = g_cumsum / Ng

        # binning and cumsum for protons
        hbins = pd.cut(hadrons, bins)
        hcount = hbins.value_counts()
        # reverse array
        hcount[:] = hcount.values[::-1].copy()
        h_cumsum = np.cumsum(hcount.values)
        eps_h = h_cumsum / Nh

        Q = eps_g / np.sqrt(eps_h)
        Q[:] = Q[::-1].copy()

        return Q, bins[1:]  # , eps_g[::-1], eps_h[::-1]

    def _hyperBinning(self, x, featsToGroupBy: list):
        """
        This function is for hyper binning with pandas. It is intended to be
        used here in order to level number of events before training the
        classifier; for more general purposes, it is the Histogram in
        utils/fitshistogram.py to be used.

        This outputs the input array grouped in as many bins as present in
        the featsToGroupBy list. This is a list of dictionaries,
        each dictionary is related to a feature (array column) to be binned.

        Beware of bins generation: since np.linspace is used, if you want log
        bins you have to make log of the input array column and pass log(max
        and min)!

        Parameters
        ----------
        x: input array to be binned

        featsToGroupBy: list(dictionary)
              list of dictionaries (see Examples for dict keys)

        Output
        ----------
        A pandas.core.groupby.DataFrameGroupBy object


        Examples
        --------

        .. code-block: python

            featsToGroupBy = [{'feat': 'feat0',
                               'maxf': max(x[:, 0]),
                               'minf': min(x[:, 0]),
                               'col': 0,
                               'nbins': 10},
                             ...
                              {'feat': 'feat1',
                               'maxf': max(x[:, 1]),
                               'minf': min(x[:, 1]),
                               'col': 1, 'nbins': 5}]


        where

        'feat' is feature name (optional);
        'maxf' and 'minf' are the range where to bin in
        'col' is feature column number in the input array
        'nbins' is the number of requested bins
        """
        dfx = pd.DataFrame(x)
        binning_list = []

        for i in range(len(featsToGroupBy)):
            feat_dict = featsToGroupBy[i]
            bins = np.linspace(feat_dict['minf'], feat_dict['maxf'],
                               feat_dict['nbins'] + 1)
            binning_list.append(pd.cut(dfx[feat_dict['col']], bins))

        groups = dfx.groupby(binning_list)

        return groups

    def level_populations(self, group_signal, group_bgd, signal_evts, bgd_evts):
        """Equalize number of entries in each bin. When doing signal -
        background separation, it is common to wrangle input data equalizing
        the number of entries for gammas and hadron in all the requested
        hyper-bins, before training the classifier. This is different from
        the sklearn train_test_split in the sense that this level the two
        populations based upon a previous binning over any feature requested.

        Parameters
        ----------
        group_signal: pandas.core.groupby.DataFrameGroupBy
            (multi dimension) histogram of signal population in pandas groups
            format
        group_bgd: pandas.core.groupby.DataFrameGroupBy
            (multi dimension) histogram of background population in pandas
             groups format
        signal_evts: np.ndarray
            array of signal events, to be equalized
        bgd_evts: np.ndarray
            array of background events, to be equalized

        Returns
        -------
        np.ndarray:
            array of gammas and array of hadrons now of the same size

        """
        logger = logging.getLogger('level_population')
        logging.basicConfig(level=logging.DEBUG)

        try:
            type(group_signal) is pd.core.groupby.DataFrameGroupBy
        except TypeError:
            raise TypeError("This function wants pandas DataFrameGroupBy "
                            "objects as group_* inputs")
        try:
            type(group_bgd) is pd.core.groupby.DataFrameGroupBy
        except TypeError:
            raise TypeError("This function wants pandas DataFrameGroupBy "
                            "objects as group_* inputs")

        # convert to pandas to use .drop()` dataframe names follow the
        # original use case gamma (dfg) vs hadrons (dfh)
        dfg = pd.DataFrame(signal_evts)
        dfh = pd.DataFrame(bgd_evts)

        # have a unique set of keys in the histograms
        s = set(group_signal.indices)
        s.update(group_bgd.indices)

        for key in s:
            if key in group_signal.indices and key in group_bgd.indices:
                # count exceeding records
                exceeding = len(group_bgd.indices[key]) - len(
                    group_signal.indices[key])

                # drop records from dataset picking exceeding number of indices
                # among those in group.indices[key]
                if exceeding > 0:
                    logger.debug('bin: %s - g: %s\th: %s\t - Removing %s '
                                 'protons',
                                 key, len(group_signal.indices[key]),
                                 len(group_bgd.indices[key]),
                                 np.abs(exceeding))

                    r_ind_list = list(
                        np.random.choice(group_bgd.indices[key],
                                         size=exceeding,
                                         replace=False)
                    )
                    dfh.drop(r_ind_list, inplace=True)
                elif exceeding < 0:
                    logger.debug('bin: %s - g: %s\th: %s\t - Removing %s '
                                 'gammas',
                                 key,
                                 len(group_signal.indices[key]),
                                 len(group_bgd.indices[key]),
                                 np.abs(exceeding))
                    r_ind_list = list(
                        np.random.choice(group_signal.indices[key],
                                         size=-exceeding,
                                         replace=False)
                    )
                    dfg.drop(r_ind_list, inplace=True)

            elif key in group_bgd.indices and key not in group_signal.indices:
                dfh.drop(group_bgd.indices[key], inplace=True)

            elif key in group_signal.indices and key not in group_bgd.indices:
                dfg.drop(group_signal.indices[key], inplace=True)

        return np.array(dfg), np.array(dfh)



import pytest
import numpy as np
from ctapipe.image.geometry_converter import (convert_geometry_hex1d_to_rect2d,
                                              convert_geometry_rect2d_back_to_hexe1d,
                                              astri_to_2d_array, array_2d_to_astri,
                                              chec_to_2d_array, array_2d_to_chec)
from ctapipe.image.hillas import hillas_parameters
from ctapipe.instrument import CameraGeometry
from ctapipe.image.toymodel import generate_2d_shower_model, make_toymodel_shower_image


cam_ids = CameraGeometry.get_known_camera_names()


@pytest.mark.parametrize("rot", [3, ])
@pytest.mark.parametrize("cam_id", cam_ids)
def test_convert_geometry(cam_id, rot):

    geom = CameraGeometry.from_name(cam_id)

    model = generate_2d_shower_model(centroid=(0.4, 0), width=0.01, length=0.03,
                                     psi="25d")

    _, image, _ = make_toymodel_shower_image(geom, model.pdf,
                                             intensity=50,
                                             nsb_level_pe=100)

    hillas_0 = hillas_parameters(geom, image)

    if geom.pix_type == 'hexagonal':
        convert_geometry_1d_to_2d = convert_geometry_hex1d_to_rect2d
        convert_geometry_back = convert_geometry_rect2d_back_to_hexe1d

        geom2d, image2d = convert_geometry_1d_to_2d(geom, image,
                                                    geom.cam_id + str(rot),
                                                    add_rot=rot)
        geom1d, image1d = convert_geometry_back(geom2d, image2d,
                                                geom.cam_id + str(rot),
                                                add_rot=rot)

    else:
        if geom.cam_id == "ASTRICam":
            convert_geometry_1d_to_2d = astri_to_2d_array
            convert_geometry_back = array_2d_to_astri
        elif geom.cam_id == "CHEC":
            convert_geometry_1d_to_2d = chec_to_2d_array
            convert_geometry_back = array_2d_to_chec
        else:
            print("camera {geom.cam_id} not implemented")
            return

        image2d = convert_geometry_1d_to_2d(image)
        image1d = convert_geometry_back(image2d)

    hillas_1 = hillas_parameters(geom, image1d)

    # if __name__ == "__main__":
    #     plot_cam(geom, geom2d, geom1d, image, image2d, image1d)
    #     plt.tight_layout()
    #     plt.pause(.1)

    assert np.abs(hillas_1.phi - hillas_0.phi).deg < 1.0
    # TODO: test other parameters


@pytest.mark.parametrize("rot", [3, ])
@pytest.mark.parametrize("cam_id", cam_ids)
def test_convert_geometry_mock(cam_id, rot):
    """here we use a different key for the back conversion to trigger the mock conversion
    """

    geom = CameraGeometry.from_name(cam_id)

    model = generate_2d_shower_model(centroid=(0.4, 0), width=0.01, length=0.03,
                                     psi="25d")

    _, image, _ = make_toymodel_shower_image(geom, model.pdf,
                                             intensity=50,
                                             nsb_level_pe=100)

    hillas_0 = hillas_parameters(geom, image)

    if geom.pix_type == 'hexagonal':
        convert_geometry_1d_to_2d = convert_geometry_hex1d_to_rect2d
        convert_geometry_back = convert_geometry_rect2d_back_to_hexe1d

        geom2d, image2d = convert_geometry_1d_to_2d(geom, image, key=None,
                                                    add_rot=rot)
        geom1d, image1d = convert_geometry_back(geom2d, image2d,
                                                "_".join([geom.cam_id,
                                                          str(rot), "mock"]),
                                                add_rot=rot)
    else:
        # originally rectangular geometries don't need a buffer and therefore no mock
        # conversion
        return

    hillas_1 = hillas_parameters(geom, image1d)
    assert np.abs(hillas_1.phi - hillas_0.phi).deg < 1.0


# def plot_cam(geom, geom2d, geom1d, image, image2d, image1d):
#     # plt.viridis()
#     plt.figure(figsize=(12, 4))
#     ax = plt.subplot(1, 3, 1)
#     CameraDisplay(geom, image=image).add_colorbar()
#     plt.subplot(1, 3, 2, sharex=ax, sharey=ax)
#     CameraDisplay(geom2d, image=image2d).add_colorbar()
#     plt.subplot(1, 3, 3, sharex=ax, sharey=ax)
#     CameraDisplay(geom1d, image=image1d).add_colorbar()
#
#
# if __name__ == "__main__":
#     import logging
#     logging.basicConfig(level=logging.DEBUG)
#     for cam_id in CameraGeometry.get_known_camera_names():
#         test_convert_geometry(cam_id, 3)
#     plt.show()



from pkg_resources import resource_filename
import os
import copy
import pytest
from ctapipe.calib.camera.calibrator import CameraCalibrator

pytest.importorskip("protozfits", minversion="0.44.3")

example_file_path = resource_filename(
    'protozfits',
    os.path.join(
        'tests',
        'resources',
        'example_10evts.fits.fz'
    )
)

FIRST_EVENT_NUMBER_IN_FILE = 97750287
ADC_SAMPLES_SHAPE = (1, 1296, 50)


def test_loop_over_events():
    from ctapipe.io.sst1meventsource import SST1MEventSource

    n_events = 5
    inputfile_reader = SST1MEventSource(
        input_url=example_file_path,
        max_events=n_events
    )

    for i, event in enumerate(inputfile_reader):
        assert event.r0.tels_with_data == {1}
        for telid in event.r0.tels_with_data:
            assert event.r0.event_id == FIRST_EVENT_NUMBER_IN_FILE + i
            assert event.r0.tel[telid].waveform.shape == ADC_SAMPLES_SHAPE

    # make sure max_events works
    assert i == n_events - 1


def test_that_event_is_not_modified_after_loop():
    from ctapipe.io.sst1meventsource import SST1MEventSource

    n_events = 2
    source = SST1MEventSource(
        input_url=example_file_path,
        max_events=n_events
    )

    for event in source:
        last_event = copy.deepcopy(event)

    # now `event` should be identical with the deepcopy of itself from
    # inside the loop.
    # Unfortunately this does not work:
    #      assert last_event == event
    # So for the moment we just compare event ids
    assert event.r0.event_id == last_event.r0.event_id


def test_is_compatible():
    from ctapipe.io.sst1meventsource import SST1MEventSource

    assert SST1MEventSource.is_compatible(example_file_path)


def test_factory_for_protozfits_file():
    from ctapipe.io.eventsourcefactory import EventSourceFactory
    from ctapipe.io.sst1meventsource import SST1MEventSource

    reader = EventSourceFactory.produce(input_url=example_file_path)
    assert isinstance(reader, SST1MEventSource)
    assert reader.input_url == example_file_path


def test_pipeline():
    from ctapipe.io.sst1meventsource import SST1MEventSource
    reader = SST1MEventSource(input_url=example_file_path, max_events=10)
    calibrator = CameraCalibrator(eventsource=reader)
    for event in reader:
        calibrator.calibrate(event)
        assert event.r0.tel.keys() == event.dl1.tel.keys()



"""
Calibrator for the R0 -> R1 data level transition.

This module handles the calibration from the R0 data level to R1. This data
level transition will be handled by the camera servers, not in the pipeline,
however the pipeline can be used as a test-bench in the comissioning stage of
the cameras.

As the R1 calibration is camera specific, each camera (and seperately the MC)
requires their own calibrator class with inherits from `CameraR1Calibrator`.
`HessioR1Calibrator` is the calibrator for the MC data obtained from readhess.
Through the use of `CameraR1CalibratorFactory`, the correct
`CameraR1Calibrator` can be obtained based on the origin (MC/Camera format)
of the data.
"""
from abc import abstractmethod
import numpy as np
from ...core import Component, Factory
from ...core.traits import Unicode
from ...io import EventSource

__all__ = [
    'NullR1Calibrator',
    'HESSIOR1Calibrator',
    'TargetIOR1Calibrator',
    'CameraR1CalibratorFactory'
]


class CameraR1Calibrator(Component):
    """
    The base R1-level calibrator. Fills the r1 container.

    The R1 calibrator performs the camera-specific R1 calibration that is
    usually performed on the raw data by the camera server. This calibrator
    exists in ctapipe for testing and prototyping purposes.

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool or None
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    kwargs
    """

    def __init__(self, config=None, tool=None, **kwargs):
        """
        Parent class for the r1 calibrators. Fills the r1 container.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool or None
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, parent=tool, **kwargs)
        self._r0_empty_warn = False

    @abstractmethod
    def calibrate(self, event):
        """
        Abstract method to be defined in child class.

        Perform the conversion from raw R0 data to R1 data
        (ADC Samples -> PE Samples), and fill the r1 container.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        """

    def check_r0_exists(self, event, telid):
        """
        Check that r0 data exists. If it does not, then do not change r1.

        This ensures that if the containers were filled from a file containing
        r0 data, it is not overwritten by non-existant data.

        Parameters
        ----------
        event : container
            A `ctapipe` event container
        telid : int
            The telescope id.

        Returns
        -------
        bool
            True if r0.tel[telid].waveform is not None, else false.
        """
        r0 = event.r0.tel[telid].waveform
        if r0 is not None:
            return True
        else:
            if not self._r0_empty_warn:
                self.log.warning("Encountered an event with no R0 data. "
                                 "R1 is unchanged in this circumstance.")
                self._r0_empty_warn = True
            return False


class NullR1Calibrator(CameraR1Calibrator):
    """
    A dummy R1 calibrator that simply fills the r1 container with the samples
    from the r0 container.

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool or None
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    kwargs
    """

    def __init__(self, config=None, tool=None, **kwargs):
        super().__init__(config, tool, **kwargs)
        self.log.info("Using NullR1Calibrator, if event source is at "
                      "the R0 level, then r1 samples will equal r0 samples")

    def calibrate(self, event):
        for telid in event.r0.tels_with_data:
            if self.check_r0_exists(event, telid):
                samples = event.r0.tel[telid].waveform
                event.r1.tel[telid].waveform = samples.astype('float32')


class HESSIOR1Calibrator(CameraR1Calibrator):
    """
    The R1 calibrator for hessio files. Fills the r1 container.

    This calibrator correctly applies the pedestal subtraction and conversion
    from counts to photoelectrons for the Monte-Carlo data.

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool or None
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    kwargs
    """

    calib_scale = 1.05
    """
    CALIB_SCALE is only relevant for MC calibration.

    CALIB_SCALE is the factor needed to transform from mean p.e. units to 
    units of the single-p.e. peak: Depends on the collection efficiency, 
    the asymmetry of the single p.e. amplitude  distribution and the 
    electronic noise added to the signals. Default value is for GCT.

    To correctly calibrate to number of photoelectron, a fresh SPE calibration
    should be applied using a SPE sim_telarray run with an 
    artificial light source.
    """
    # TODO: Handle calib_scale differently per simlated telescope

    def calibrate(self, event):
        if event.meta['origin'] != 'hessio':
            raise ValueError('Using HESSIOR1Calibrator to calibrate a '
                             'non-hessio event.')

        for telid in event.r0.tels_with_data:
            if self.check_r0_exists(event, telid):
                samples = event.r0.tel[telid].waveform
                n_samples = samples.shape[2]
                ped = event.mc.tel[telid].pedestal / n_samples
                gain = event.mc.tel[telid].dc_to_pe * self.calib_scale
                calibrated = (samples - ped[..., None]) * gain[..., None]
                event.r1.tel[telid].waveform = calibrated


class TargetIOR1Calibrator(CameraR1Calibrator):

    pedestal_path = Unicode(
        '',
        allow_none=True,
        help='Path to the TargetCalib pedestal file'
    ).tag(config=True)
    tf_path = Unicode(
        '',
        allow_none=True,
        help='Path to the TargetCalib Transfer Function file'
    ).tag(config=True)
    pe_path = Unicode(
        '',
        allow_none=True,
        help='Path to the TargetCalib pe conversion file'
    ).tag(config=True)
    ff_path = Unicode(
        '',
        allow_none=True,
        help='Path to a TargetCalib flat field file'
    ).tag(config=True)

    def __init__(self, config=None, tool=None, **kwargs):
        """
        The R1 calibrator for targetio files (i.e. files containing data
        taken with a TARGET module, such as with CHEC)

        Fills the r1 container.

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs
        """
        super().__init__(config=config, tool=tool, **kwargs)
        try:
            import target_calib
        except ImportError:
            msg = ("Cannot find target_calib module, please follow "
                   "installation instructions from https://forge.in2p3.fr/"
                   "projects/gct/wiki/Installing_CHEC_Software")
            self.log.error(msg)
            raise

        self._r1_wf = None
        self.tc = target_calib
        self.calibrator = None
        self.telid = 0

        self._load_calib()

    def calibrate(self, event):
        """
        Placeholder function to satisfy abstract parent, this is overloaded by
        either fake_calibrate or real_calibrate.
        """
        pass

    def _load_calib(self):
        """
        If a pedestal file has been supplied, create a target_calib
        Calibrator object. If it hasn't then point calibrate to
        fake_calibrate, where nothing is done to the waveform.
        """
        if self.pedestal_path:
            self.calibrator = self.tc.Calibrator(self.pedestal_path,
                                                 self.tf_path,
                                                 [self.pe_path, self.ff_path])
            self.calibrate = self.real_calibrate
        else:
            self.log.warning("No pedestal path supplied, "
                             "r1 samples will equal r0 samples.")
            self.calibrate = self.fake_calibrate

    def fake_calibrate(self, event):
        """
        Don't perform any calibration on the waveforms, just fill the
        R1 container.

        Parameters
        ----------
        event : `ctapipe` event-container
        """
        if event.meta['origin'] != 'targetio':
            raise ValueError('Using TargetioR1Calibrator to calibrate a '
                             'non-targetio event.')

        if self.check_r0_exists(event, self.telid):
            samples = event.r0.tel[self.telid].waveform
            event.r1.tel[self.telid].waveform = samples.astype('float32')

    def real_calibrate(self, event):
        """
        Apply the R1 calibration defined in target_calib and fill the
        R1 container.

        Parameters
        ----------
        event : `ctapipe` event-container
        """
        if event.meta['origin'] != 'targetio':
            raise ValueError('Using TargetioR1Calibrator to calibrate a '
                             'non-targetio event.')

        if self.check_r0_exists(event, self.telid):
            samples = event.r0.tel[self.telid].waveform
            if self._r1_wf is None:
                self._r1_wf = np.zeros(samples.shape, dtype=np.float32)
            fci = event.targetio.tel[self.telid].first_cell_ids
            r1 = event.r1.tel[self.telid].waveform[0]
            self.calibrator.ApplyEvent(samples[0], fci, self._r1_wf[0])
            event.r1.tel[self.telid].waveform = self._r1_wf


class CameraR1CalibratorFactory(Factory):
    """
    The R1 calibrator `ctapipe.core.factory.Factory`. This
    `ctapipe.core.factory.Factory` allows the correct
    `CameraR1Calibrator` to be obtained for the data investigated.

    Additional filepaths are required by some cameras for R1 calibration. Due
    to the current inplementation of `ctapipe.core.factory.Factory`, every
    trait that could
    possibly be required for a child `ctapipe.core.component.Component` of
    `CameraR1Calibrator` is
    included in this `ctapipe.core.factory.Factory`. The
    `CameraR1Calibrator` specific to a
    camera type should then define how/if that filepath should be used. The
    format of the file is not restricted, and the file can be read from inside
    ctapipe, or can call a different library created by the camera teams for
    the calibration of their camera.
    """
    base = CameraR1Calibrator
    custom_product_help = ('R1 Calibrator to use. If None then a '
                           'calibrator will either be selected based on the '
                           'supplied EventSource, or will default to '
                           '"NullR1Calibrator".')

    def __init__(self, config=None, tool=None, eventsource=None, **kwargs):
        """
        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        eventsource : ctapipe.io.eventsource.EventSource
            EventSource that is being used to read the events. The EventSource
            contains information (such as metadata or inst) which indicates
            the appropriate R1Calibrator to use.
        kwargs

        """

        super().__init__(config, tool, **kwargs)
        if eventsource and not issubclass(type(eventsource), EventSource):
            raise TypeError(
                "eventsource must be a ctapipe.io.eventsource.EventSource"
            )
        self.eventsource = eventsource

    def _get_product_name(self):
        try:
            return super()._get_product_name()
        except AttributeError:
            es = self.eventsource
            if es:
                if es.metadata['is_simulation']:
                    return 'HESSIOR1Calibrator'
                elif es.__class__.__name__ == "TargetIOEventSource":
                    return 'TargetIOR1Calibrator'
            return 'NullR1Calibrator'



# skip these tests if matplotlib can't be imported
import pytest

plt = pytest.importorskip("matplotlib.pyplot")

from ctapipe.instrument import (CameraGeometry, SubarrayDescription,
                                TelescopeDescription)
from ctapipe.io.containers import HillasParametersContainer
from numpy import ones
from astropy import units as u


def test_camera_display_single():
    """ test CameraDisplay functionality """
    from ..mpl_camera import CameraDisplay

    geom = CameraGeometry.from_name("LSTCam")
    disp = CameraDisplay(geom)
    image = ones(len(geom.pix_x), dtype=float)
    disp.image = image
    disp.add_colorbar()
    disp.cmap = 'nipy_spectral'
    disp.set_limits_minmax(0, 10)
    disp.set_limits_percent(95)
    disp.enable_pixel_picker()
    disp.highlight_pixels([1, 2, 3, 4, 5])
    disp.norm = 'log'
    disp.norm = 'symlog'
    disp.cmap = 'rainbow'

    with pytest.raises(ValueError):
        disp.image = ones(10)

    with pytest.raises(ValueError):
        disp.add_colorbar()

    disp.add_ellipse(centroid=(0, 0), width=0.1, length=0.1, angle=0.1)
    disp.clear_overlays()


def test_camera_display_multiple():
    """ create a figure with 2 subplots, each with a CameraDisplay """
    from ..mpl_camera import CameraDisplay

    geom = CameraGeometry.from_name("LSTCam")
    fig, ax = plt.subplots(2, 1)

    d1 = CameraDisplay(geom, ax=ax[0])
    d2 = CameraDisplay(geom, ax=ax[1])

    image = ones(len(geom.pix_x), dtype=float)
    d1.image = image
    d2.image = image


def test_array_display():
    from ctapipe.visualization.mpl_array import ArrayDisplay

    # build a test subarray:
    tels = dict()
    tel_pos = dict()
    for ii, pos in enumerate([[0, 0, 0], [100, 0, 0], [-100, 0, 0]] * u.m):
        tels[ii + 1] = TelescopeDescription.from_name("MST", "NectarCam")
        tel_pos[ii + 1] = pos

    sub = SubarrayDescription(
        name="TestSubarray",
        tel_positions=tel_pos,
        tel_descriptions=tels
    )

    ad = ArrayDisplay(sub)
    ad.set_vector_rho_phi(1 * u.m, 90 * u.deg)

    # try setting a value
    vals = ones(sub.num_tels)
    ad.values = vals

    assert (vals == ad.values).all()

    # test using hillas params:
    hillas_dict = {
        1: HillasParametersContainer(length=1.0 * u.m, phi=90 * u.deg),
        2: HillasParametersContainer(length=200 * u.cm, phi="95deg"),
    }
    ad.set_vector_hillas(hillas_dict)

    ad.add_labels()
    ad.remove_labels()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Coordinates.
"""


from .coordinate_transformations import (pixel_position_to_direction,
                                         alt_to_theta, az_to_phi,
                                         transform_pixel_position)


from .angular_frames import *
from .ground_frames import *


from astropy.utils import iers
iers.conf.auto_download = True  # auto-fetch updates to IERS_A table



from astropy import units as u
from astropy.coordinates import Angle
from astropy.time import Time
from ctapipe.io.eventsource import EventSource
from ctapipe.io.containers import DataContainer
from ctapipe.instrument import TelescopeDescription, SubarrayDescription
import gzip
import struct

__all__ = ['HESSIOEventSource']


class HESSIOEventSource(EventSource):
    """
    EventSource for the hessio file format.

    This class utilises `pyhessio` to read the hessio file, and stores the
    information into the event containers.
    """
    _count = 0

    def __init__(self, config=None, tool=None, **kwargs):
        super().__init__(config=config, tool=tool, **kwargs)

        try:
            import pyhessio
        except ImportError:
            msg = "The `pyhessio` python module is required to access MC data"
            self.log.error(msg)
            raise

        self.pyhessio = pyhessio

        if HESSIOEventSource._count > 0:
            self.log.warn("Only one pyhessio event_source allowed at a time. "
                          "Previous hessio file will be closed.")
            self.pyhessio.close_file()
        HESSIOEventSource._count += 1

        self.metadata['is_simulation'] = True

    @staticmethod
    def is_compatible(file_path):
        # read the first 4 bytes
        with open(file_path, 'rb') as f:
            marker_bytes = f.read(4)
        # if file is gzip, read the first 4 bytes with gzip again
        if marker_bytes[0] == 0x1f and marker_bytes[1] == 0x8b:
            with gzip.open(file_path, 'rb') as f:
                marker_bytes = f.read(4)
        # check for the simtel magic marker
        int_marker, = struct.unpack('I', marker_bytes)
        return int_marker == 3558836791 or int_marker == 931798996

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.pyhessio.close_file()

    def _generator(self):
        with self.pyhessio.open_hessio(self.input_url) as file:
            # the container is initialized once, and data is replaced within
            # it after each yield
            counter = 0
            eventstream = file.move_to_next_event()
            data = DataContainer()
            data.meta['origin'] = "hessio"

            # some hessio_event_source specific parameters
            data.meta['input_url'] = self.input_url
            data.meta['max_events'] = self.max_events

            for event_id in eventstream:

                if counter == 0:
                    # subarray info is only available when an event is loaded,
                    # so load it on the first event.
                    data.inst.subarray = self._build_subarray_info(file)

                obs_id = file.get_run_number()
                tels_with_data = set(file.get_teldata_list())
                data.count = counter
                data.r0.obs_id = obs_id
                data.r0.event_id = event_id
                data.r0.tels_with_data = tels_with_data
                data.r1.obs_id = obs_id
                data.r1.event_id = event_id
                data.r1.tels_with_data = tels_with_data
                data.dl0.obs_id = obs_id
                data.dl0.event_id = event_id
                data.dl0.tels_with_data = tels_with_data

                # handle telescope filtering by taking the intersection of
                # tels_with_data and allowed_tels
                if len(self.allowed_tels) > 0:
                    selected = tels_with_data & self.allowed_tels
                    if len(selected) == 0:
                        continue  # skip event
                    data.r0.tels_with_data = selected
                    data.r1.tels_with_data = selected
                    data.dl0.tels_with_data = selected

                data.trig.tels_with_trigger = (file.
                                               get_central_event_teltrg_list())
                time_s, time_ns = file.get_central_event_gps_time()
                data.trig.gps_time = Time(time_s * u.s, time_ns * u.ns,
                                          format='unix', scale='utc')
                data.mc.energy = file.get_mc_shower_energy() * u.TeV
                data.mc.alt = Angle(file.get_mc_shower_altitude(), u.rad)
                data.mc.az = Angle(file.get_mc_shower_azimuth(), u.rad)
                data.mc.core_x = file.get_mc_event_xcore() * u.m
                data.mc.core_y = file.get_mc_event_ycore() * u.m
                first_int = file.get_mc_shower_h_first_int() * u.m
                data.mc.h_first_int = first_int
                data.mc.shower_primary_id = file.get_mc_shower_primary_id()

                # mc run header data
                data.mcheader.run_array_direction = (
                    file.get_mc_run_array_direction()
                ) * u.rad

                # this should be done in a nicer way to not re-allocate the
                # data each time (right now it's just deleted and garbage
                # collected)

                data.r0.tel.clear()
                data.r1.tel.clear()
                data.dl0.tel.clear()
                data.dl1.tel.clear()
                data.mc.tel.clear()  # clear the previous telescopes

                for tel_id in tels_with_data:

                    # event.mc.tel[tel_id] = MCCameraContainer()

                    data.mc.tel[tel_id].dc_to_pe = file.get_calibration(tel_id)
                    data.mc.tel[tel_id].pedestal = file.get_pedestal(tel_id)
                    data.r0.tel[tel_id].waveform = (file.
                                                    get_adc_sample(tel_id))
                    if data.r0.tel[tel_id].waveform.size == 0:
                        # To handle ASTRI and dst files
                        data.r0.tel[tel_id].waveform = (file.
                                                        get_adc_sum(tel_id)[..., None])
                    data.r0.tel[tel_id].image = file.get_adc_sum(tel_id)
                    data.mc.tel[tel_id].reference_pulse_shape = (file.
                                                                 get_ref_shapes(tel_id))

                    nsamples = file.get_event_num_samples(tel_id)
                    if nsamples <= 0:
                        nsamples = 1
                    data.r0.tel[tel_id].num_samples = nsamples

                    # load the data per telescope/pixel
                    hessio_mc_npe = file.get_mc_number_photon_electron(tel_id)
                    data.mc.tel[tel_id].photo_electron_image = hessio_mc_npe
                    data.mc.tel[tel_id].meta['refstep'] = (file.
                                                           get_ref_step(tel_id))
                    data.mc.tel[tel_id].time_slice = (file.
                                                      get_time_slice(tel_id))
                    data.mc.tel[tel_id].azimuth_raw = (file.
                                                       get_azimuth_raw(tel_id))
                    data.mc.tel[tel_id].altitude_raw = (file.
                                                        get_altitude_raw(tel_id))
                    data.mc.tel[tel_id].azimuth_cor = (file.
                                                       get_azimuth_cor(tel_id))
                    data.mc.tel[tel_id].altitude_cor = (file.
                                                        get_altitude_cor(tel_id))
                yield data
                counter += 1

        return

    def _build_subarray_info(self, file):
        """
        constructs a SubarrayDescription object from the info in an
        EventIO/HESSSIO file

        Parameters
        ----------
        file: HessioFile
            The open pyhessio file

        Returns
        -------
        SubarrayDescription :
            instrumental information
        """
        telescope_ids = list(file.get_telescope_ids())
        subarray = SubarrayDescription("MonteCarloArray")

        for tel_id in telescope_ids:
            try:

                pix_pos = file.get_pixel_position(tel_id) * u.m
                foclen = file.get_optical_foclen(tel_id) * u.m
                mirror_area = file.get_mirror_area(tel_id) * u.m ** 2
                num_tiles = file.get_mirror_number(tel_id)
                tel_pos = file.get_telescope_position(tel_id) * u.m

                tel = TelescopeDescription.guess(*pix_pos,
                                                 equivalent_focal_length=foclen)
                tel.optics.mirror_area = mirror_area
                tel.optics.num_mirror_tiles = num_tiles
                subarray.tels[tel_id] = tel
                subarray.positions[tel_id] = tel_pos

            except self.pyhessio.HessioGeneralError:
                pass

        return subarray



from itertools import cycle

import numpy as np
from astropy import units as u
from astropy.coordinates import Angle
from matplotlib import pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.lines import Line2D
from matplotlib.patches import Circle

from ctapipe.coordinates import GroundFrame
from ctapipe.visualization.mpl_camera import polar_to_cart


class ArrayDisplay:
    """
    Display a top-town view of a telescope array.

    This can be used in two ways: by default, you get a display of all
    telescopes in the subarray, colored by telescope type, however you can
    also color the telescopes by a value (like trigger pattern, or some other
    scalar per-telescope parameter). To set the color value, simply set the
    `value` attribute, and the fill color will be updated with the value. You
    might want to set the border color to zero to avoid confusion between the
    telescope type color and the value color (
    `array_disp.telescope.set_linewidth(0)`)

    To display a vector field over the telescope positions, e.g. for
    reconstruction, call `set_uv()` to set cartesian vectors, or `set_r_phi()`
    to set polar coordinate vectors.  These both take an array of length
    N_tels, or a single value.


    Parameters
    ----------
    subarray: ctapipe.instrument.SubarrayDescription
        the array layout to display
    axes: matplotlib.axes.Axes
        matplotlib axes to plot on, or None to use current one
    title: str
        title of array plot
    tel_scale: float
        scaling between telescope mirror radius in m to displayed size
    autoupdate: bool
        redraw when the input changes
    radius: Union[float, list, None]
        set telescope radius to value, list/array of values. If None, radius
        is taken from the telescope's mirror size.
    """

    def __init__(self, subarray, axes=None, autoupdate=True,
                 tel_scale=2.0, alpha=0.7, title=None,
                 radius=None, frame=GroundFrame()):

        self.frame = frame
        self.subarray = subarray

        # get the telescope positions. If a new frame is set, this will
        # transform to the new frame.
        self.tel_coords = subarray.tel_coords.transform_to(frame)

        # set up colors per telescope type
        tel_types = [str(tel) for tel in subarray.tels.values()]
        if radius is None:
            # set radius to the mirror radius (so big tels appear big)
            radius = [np.sqrt(tel.optics.mirror_area.to("m2").value) * tel_scale
                      for tel in subarray.tel.values()]

        if title is None:
            title = subarray.name

        # get default matplotlib color cycle (depends on the current style)
        color_cycle = cycle(plt.rcParams['axes.prop_cycle'].by_key()['color'])

        # map a color to each telescope type:
        tel_type_to_color = {}
        for tel_type in list(set(tel_types)):
            tel_type_to_color[tel_type] = next(color_cycle)

        tel_color = [tel_type_to_color[ttype] for ttype in tel_types]

        patches = []
        for x, y, r, c in zip(list(self.tel_coords.x.value),
                              list(self.tel_coords.y.value),
                              list(radius),
                              tel_color):
            patches.append(
                Circle(
                    xy=(x, y),
                    radius=r,
                    fill=True,
                    color=c,
                    alpha=alpha,
                )
            )

        # build the legend:
        legend_elements = []
        for ttype in list(set(tel_types)):
            color = tel_type_to_color[ttype]
            legend_elements.append(
                Line2D([0], [0], marker='o', color=color,
                       label=ttype, markersize=10, alpha=alpha,
                       linewidth=0)
            )
        plt.legend(handles=legend_elements)

        self.tel_colors = tel_color
        self.autoupdate = autoupdate
        self.telescopes = PatchCollection(patches, match_original=True)
        self.telescopes.set_linewidth(2.0)

        self.axes = axes or plt.gca()
        self.axes.add_collection(self.telescopes)
        self.axes.set_aspect(1.0)
        self.axes.set_title(title)
        self._labels = []
        self._quiver = None
        self.axes.autoscale_view()

    @property
    def values(self):
        """An array containing a value per telescope"""
        return self.telescopes.get_array()

    @values.setter
    def values(self, values):
        """ set the telescope colors to display  """
        self.telescopes.set_array(values)
        self._update()

    def set_vector_uv(self, u, v, c=None, **kwargs):
        """ sets the vector field U,V and color for all telescopes

        Parameters
        ----------
        u: array[num_tels]
            x-component of direction vector
        v: array[num_tels]
            y-component of direction vector
        kwargs:
            extra args passed to plt.quiver(), ignored on subsequent updates
        """
        if c is None:
            c = self.tel_colors

        if self._quiver is None:
            coords = self.tel_coords
            self._quiver = self.axes.quiver(
                coords.x, coords.y,
                u, v,
                color=c,
                **kwargs
            )
        else:
            self._quiver.set_UVC(u, v)

    def set_vector_rho_phi(self, rho, phi, c=None, **kwargs):
        """sets the vector field using R, Phi for each telescope

        Parameters
        ----------
        rho: float or array[float]
            vector magnitude for each telescope
        phi: array[Angle]
            vector angle for each telescope
        c: color or list of colors
            vector color for each telescope (or one for all)
        """
        phi = Angle(phi).rad
        u, v = polar_to_cart(rho, phi)
        self.set_vector_uv(u, v, c=c, **kwargs)

    def set_vector_hillas(self, hillas_dict, angle_offset=180 * u.deg):
        """
        helper function to set the vector angle and length from a set of
        Hillas parameters.

        Parameters
        ----------
        hillas_dict: Dict[int, HillasParametersContainer]
            mapping of tel_id to Hillas parameters
        """

        rho = np.zeros(self.subarray.num_tels) * u.m
        phi = np.zeros(self.subarray.num_tels) * u.deg

        for tel_id, params in hillas_dict.items():
            idx = self.subarray.tel_indices[tel_id]
            rho[idx] = 1.0 * u.m  # params.length
            phi[idx] = Angle(params.phi) + Angle(angle_offset)

        self.set_vector_rho_phi(rho=rho, phi=phi)

    def add_labels(self):
        px = self.tel_coords.x.value
        py = self.tel_coords.y.value
        for tel, x, y in zip(self.subarray.tels, px, py):
            name = str(tel)
            lab = self.axes.text(x, y, name, fontsize=8)
            self._labels.append(lab)

    def remove_labels(self):
        for lab in self._labels:
            lab.remove()
        self._labels = []

    def _update(self):
        """ signal a redraw if necessary """
        if self.autoupdate:
            plt.draw()

    def background_contour(self, x, y, background, **kwargs):
        """
        Draw image contours in background of the display, useful when likelihood fitting

        Parameters
        ----------
        x: ndarray
            array of image X coordinates
        y: ndarray
            array of image Y coordinates
        background: ndarray
            Array of image to use in background
        kwargs: key=value
            any style keywords to pass to matplotlib
        """

        # use zorder to ensure the contours appear under the telescopes.
        self.axes.contour(x, y, background, zorder=0, **kwargs)



from copy import deepcopy
from numpy.testing import assert_allclose
from ctapipe.calib.camera import (
    CameraCalibrator,
    HESSIOR1Calibrator,
    NullR1Calibrator
)
from ctapipe.image.charge_extractors import LocalPeakIntegrator
from ctapipe.io import HESSIOEventSource
from ctapipe.utils import get_dataset_path


def test_camera_calibrator(test_event):
    event = deepcopy(test_event) # so we don't modify the test event
    telid = 11

    calibrator = CameraCalibrator(r1_product="HESSIOR1Calibrator")

    calibrator.calibrate(event)
    image = event.dl1.tel[telid].image
    assert_allclose(image[0, 0], -2.216, 1e-3)


def test_manual_r1():
    calibrator = CameraCalibrator(r1_product="HESSIOR1Calibrator")
    assert isinstance(calibrator.r1, HESSIOR1Calibrator)


def test_manual_extractor():
    calibrator = CameraCalibrator(extractor_product="LocalPeakIntegrator")
    assert isinstance(calibrator.dl1.extractor, LocalPeakIntegrator)


def test_eventsource_r1():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    eventsource = HESSIOEventSource(input_url=dataset)
    calibrator = CameraCalibrator(eventsource=eventsource)
    assert isinstance(calibrator.r1, HESSIOR1Calibrator)


def test_eventsource_override_r1():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    eventsource = HESSIOEventSource(input_url=dataset)
    calibrator = CameraCalibrator(
        eventsource=eventsource,
        r1_product="NullR1Calibrator"
    )
    assert isinstance(calibrator.r1, NullR1Calibrator)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import json
import logging
import os
import re

import yaml
from astropy.table import Table
from astropy.utils.decorators import deprecated
from pkg_resources import resource_listdir

from ..core import Provenance

logger = logging.getLogger(__name__)

try:
    import ctapipe_resources
except ImportError:
    raise RuntimeError("Please install the 'ctapipe-extra' package, "
                       "which contains the ctapipe_resources module "
                       "needed by ctapipe. (conda install ctapipe-extra)")

__all__ = ['get_dataset_path', 'find_in_path', 'find_all_matching_datasets']


def get_searchpath_dirs(searchpath=os.getenv("CTAPIPE_SVC_PATH")):
    """ returns a list of dirs in specified searchpath"""
    if searchpath == "" or searchpath is None:
        return []
    return os.path.expandvars(searchpath).split(':')


def find_all_matching_datasets(pattern,
                               searchpath=None,
                               regexp_group=None):
    """
    Returns a list of resource names (or substrings) matching the given 
    pattern, searching first in searchpath (a colon-separated list of 
    directories) and then in the ctapipe_resources module)

    Parameters
    ----------
    pattern: str
       regular expression to use for matching
    searchpath: str
       colon-seprated list of directories in which to search, defaulting to 
       CTAPIPE_SVC_PATH environment variable
    regexp_group: int
       if not None, return the regular expression group indicated (assuming 
       pattern has a group specifier in it)

    Returns
    -------
    list(str):
       resources names, use get_dataset_path() to retrieve the full filename
    """
    results = set()

    if searchpath is None:
        searchpath = os.getenv("CTAPIPE_SVC_PATH")

    # first check search path
    if searchpath is not None:
        for path in get_searchpath_dirs(searchpath):
            if os.path.exists(path):
                for filename in os.listdir(path):
                    match = re.match(pattern, filename)
                    if match:
                        if regexp_group is not None:
                            results.add(match.group(regexp_group))
                        else:
                            results.add(filename)

    # then check resources module
    for resource in resource_listdir('ctapipe_resources', ''):
        match = re.match(pattern, resource)
        if match:
            if regexp_group is not None:
                results.add(match.group(regexp_group))
            else:
                results.add(resource)

    return list(results)


def find_in_path(filename, searchpath):
    """
    Search in searchpath for filename, returning full path.

    Parameters
    ----------
    searchpath: str
        colon-separated list of directories (like PATH)
    filename:
        filename to find

    Returns
    -------
    full path to file if found, None otherwise

    """

    for directory in get_searchpath_dirs(searchpath):
        pathname = os.path.join(directory, filename)
        if os.path.exists(pathname):
            return pathname

    return None


def get_dataset_path(filename):
    """
    Returns the full file path to an auxiliary dataset needed by 
    ctapipe, given the dataset's full name (filename with no directory).

    This will first search for the file in directories listed in 
    tne environment variable CTAPIPE_SVC_PATH (if set), and if not found,  
    will look in the ctapipe_resources module 
    (installed with the ctapipe-extra package), which contains the defaults.

    Parameters
    ----------
    filename: str
        name of dataset to fetch

    Returns
    -------
    string with full path to the given dataset
    """
    searchpath = os.getenv("CTAPIPE_SVC_PATH")

    if searchpath:
        filepath = find_in_path(filename=filename, searchpath=searchpath)
        if filepath:
            return filepath

    logger.debug("Resource '{}' not found in CTAPIPE_SVC_PATH, looking in "
                 "ctapipe_resources...".format(filename))

    return ctapipe_resources.get(filename)


@deprecated(0.6, "use get_dataset_path instead")
def get_dataset(filename):
    return get_dataset_path(filename)


def get_table_dataset(table_name, role='resource', **kwargs):
    """
    get a tabular dataset as an `astropy.table.Table` object

    Parameters
    ----------
    table_name: str
        base name of table, without file extension
    role: str
        should be set to the CTA data hierarchy name when possible (e.g.
        dl1.sub.svc.arraylayout). This will be recorded in the provenance
        system.
    kwargs:
        extra arguments to pass to Table.read()

    Returns
    -------
    Table
    """

    # a mapping of types (keys) to any extra keyword args needed for
    # table.read()
    types_to_try = {
        '.fits.gz': {},
        '.fits': {},
        '.ecsv': dict(format='ascii.ecsv'),
        '.ecsv.txt': dict(format='ascii.ecsv'),
    }

    for table_type in types_to_try:
        filename = table_name + table_type
        try:
            fullname = get_dataset_path(filename)
            if fullname:
                args = types_to_try[table_type]
                args.update(kwargs)
                table = Table.read(fullname, **args)
                Provenance().add_input_file(fullname, role)
                return table
        except FileNotFoundError:
            pass

    raise FileNotFoundError("couldn't locate table: {}[{}]".format(
        table_name, ', '.join(types_to_try)))


def get_structured_dataset(basename, role='resource', **kwargs):
    """
    find and return a YAML or JSON dataset as a dictionary

    Parameters
    ----------
    basename: str
        base-name (without extension) of the dataset
    role: str
        provenence role
    kwargs: dict
        any other arguments to pass on to yaml or json loader.

    Returns
    -------
    dict:
       dictionary of data in the file
    """

    # a mapping of types (keys) to any extra keyword args needed for
    # table.read()
    types_to_try = {
        '.yaml': {},
        '.yml': {},
        '.json': {},
    }

    for data_type in types_to_try:
        filename = basename + data_type
        try:
            fullname = get_dataset_path(filename)
            if fullname:
                args = types_to_try[data_type]
                args.update(kwargs)

                with open(fullname) as infile:
                    if data_type == '.yaml' or data_type == '.yml':
                        dataset = yaml.load(infile, **args)
                    elif data_type == '.json':
                        dataset = json.load(infile, **args)

                Provenance().add_input_file(fullname, role)
                return dataset
        except FileNotFoundError:
            pass

    raise FileNotFoundError("couldn't locate structed dataset: {}[{}]".format(
        basename, ', '.join(types_to_try)))


@deprecated("ctapipe-0.5", alternative='get_dataset()')
def get_path(filename):
    return get_dataset_path(filename)


if __name__ == '__main__':
    get_table_dataset("NectarCam.camgeom")



from ctapipe.core.factory import Factory
from ctapipe.io.eventsource import EventSource

# EventFileReader imports so that EventFileReaderFactory can see them
# (they need to exist in the global namespace)
import ctapipe.io.hessioeventsource
from . import sst1meventsource
from . import nectarcameventsource
import ctapipe.io.targetioeventsource


__all__ = ['EventSourceFactory', 'event_source']


class EventSourceFactory(Factory):
    """
    The `EventSource` `ctapipe.core.factory.Factory`. This
    `ctapipe.core.factory.Factory` allows the correct
    `EventSource` to be obtained for the event file being read.

    This factory tests each EventSource by calling
    `EventSource.check_file_compatibility` to see which `EventSource`
    is compatible with the file.

    Using `EventFileReaderFactory` in a script allows it to be compatible with
    any file format that has an `EventSource` defined.

    To use within a `ctapipe.core.tool.Tool`:

    >>> event_source = EventSourceFactory.produce(config=self.config, tool=self)

    Parameters
    ----------
    config : traitlets.loader.Config
        Configuration specified by config file or cmdline arguments.
        Used to set traitlet values.
        Set to None if no configuration to pass.
    tool : ctapipe.core.Tool or None
        Tool executable that is calling this component.
        Passes the correct logger to the component.
        Set to None if no Tool to pass.
    kwargs

    Attributes
    ----------
    product : traitlets.CaselessStrEnum
        A string with the `EventSource.__name__` of the reader you want to
        use. If left blank, `EventSource.check_file_compatibility` will be
        used to find a compatible event_source.
    """
    base = EventSource
    custom_product_help = ('EventSource to use. If None then a reader will '
                           'be chosen based on the input_url')
    input_url = None  # Instanced as a traitlet by FactoryMeta

    def _get_product_name(self):
        try:
            return super()._get_product_name()
        except AttributeError:
            if self.input_url is None:
                raise ValueError("Please specify an input_url for event file")
            try:
                for subclass in self.subclasses:
                    if subclass.is_compatible(self.input_url):
                        return subclass.__name__
                raise ValueError
            except ValueError:
                self.log.exception("Cannot find compatible EventSource "
                                   "for: {}".format(self.input_url))
                raise


def event_source(input_url, config=None, parent=None, **kwargs):
    """
    Helper function to quickly construct an `EventSourceFactory` and produce
    an `EventSource`. This may be used in small scripts and demos for
    simplicity. In a `ctapipe.core.Tool` class, a `EventSourceFactory` should
    be manually constructed, so that the configuration info is correctly
    passed in.

    Examples
    --------
    >>> with event_source(url) as source:
    >>>    for event in source:
    >>>         print(event.r0.event_id)

    Parameters
    ----------
    input_url: str
        filename or URL pointing to an event file.

    Returns
    -------
    EventSource:
        a properly constructed `EventSource` subclass, depending on the
        input filename.
    """

    reader = EventSourceFactory.produce(config, parent,
                                        input_url=input_url,
                                        **kwargs)

    return reader



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
EventSource for SST1M/digicam protobuf-fits.fz-files.

Needs protozfits v0.44.3 from github.com/cta-sst-1m/protozfitsreader
"""
import numpy as np
from .eventsource import EventSource
from .containers import SST1MDataContainer
from ..instrument import TelescopeDescription

__all__ = ['SST1MEventSource']


class SST1MEventSource(EventSource):

    def __init__(self, config=None, tool=None, **kwargs):
        super().__init__(config=config, tool=tool, **kwargs)
        from protozfits import SimpleFile
        self.file = SimpleFile(self.input_url)
        # TODO: Correct pixel ordering
        self._tel_desc = TelescopeDescription.from_name(
            optics_name='SST-1M',
            camera_name='DigiCam'
        )

    def _generator(self):
        pixel_sort_ids = None

        for count, event in enumerate(self.file.Events):
            if pixel_sort_ids is None:
                pixel_indices = event.hiGain.waveforms.pixelsIndices
                pixel_sort_ids = np.argsort(pixel_indices)
                self.n_pixels = len(pixel_sort_ids)
            telid = event.telescopeID
            data = SST1MDataContainer()
            data.count = count

            data.inst.subarray.tels[telid] = self._tel_desc

            # Data level Containers
            data.r0.obs_id = -1
            data.r0.event_id = event.eventNumber
            data.r0.tels_with_data = {telid}
            data.r1.obs_id = -1
            data.r1.event_id = event.eventNumber
            data.r1.tels_with_data = {telid}
            data.dl0.obs_id = -1
            data.dl0.event_id = event.eventNumber
            data.dl0.tels_with_data = {telid}

            # R0CameraContainer
            camera_time = event.local_time_sec * 1E9 + event.local_time_nanosec
            samples = event.hiGain.waveforms.samples.reshape(self.n_pixels, -1)
            data.r0.tel[telid].trigger_time = camera_time
            data.r0.tel[telid].trigger_type = event.event_type
            data.r0.tel[telid].waveform = samples[pixel_sort_ids][None, :]
            data.r0.tel[telid].num_samples = samples.shape[-1]

            # SST1MContainer
            data.sst1m.tels_with_data = {telid}

            # SST1MCameraContainer
            digicam_baseline = event.hiGain.waveforms.baselines[pixel_sort_ids]
            gps_time = (event.trig.timeSec * 1E9 + event.trig.timeNanoSec)
            container = data.sst1m.tel[telid]
            container.pixel_flags = event.pixels_flags[pixel_sort_ids]
            container.digicam_baseline = digicam_baseline
            container.local_camera_clock = camera_time
            container.gps_time = gps_time
            container.camera_event_type = event.event_type
            container.array_event_type = event.eventType
            container.trigger_input_traces = event.trigger_input_traces
            container.trigger_output_patch7 = event.trigger_output_patch7
            container.trigger_output_patch19 = event.trigger_output_patch19

            yield data

    @staticmethod
    def is_compatible(file_path):
        from astropy.io import fits
        try:
            h = fits.open(file_path)[1].header
            ttypes = [h[x] for x in h.keys() if 'TTYPE' in x]
        except OSError:
            # not even a fits file
            return False
        except IndexError:
            # A fits file of a different format
            return False

        is_protobuf_zfits_file = (
            (h['XTENSION'] == 'BINTABLE') and
            (h['EXTNAME'] == 'Events') and
            (h['ZTABLE'] is True) and
            (h['ORIGIN'] == 'CTA') and
            (h['PBFHEAD'] == 'DataModel.CameraEvent')
        )
        is_sst1m_file = 'trigger_input_traces' in ttypes

        return is_protobuf_zfits_file & is_sst1m_file



__author__ = "@jeremiedecock"

import numpy as np


def chec_transformation_map():
    # By default, pixels map to the last element of input_img_ext (i.e. NaN)
    img_map = np.full([8 * 6, 8 * 6], -1, dtype=int)

    # Map values
    img_map[:8, 8:-8] = np.arange(8 * 8 * 4).reshape([8, 8 * 4])
    img_map[8:40, :] = np.arange(32 * 48).reshape([32, 48]) + 256
    img_map[-8:, 8:-8] = np.arange(8 * 8 * 4).reshape([8, 8 * 4]) + 1792

    return img_map


def chec_to_2d_array(input_img, img_map=chec_transformation_map()):
    """
    Convert images comming form "CHEC" cameras in order to get regular 2D
    "rectangular" images directly usable with most image processing tools.

    Parameters
    ----------
    input_img : numpy.array
        The image to convert

    Returns
    -------
    A numpy.array containing the cropped image.
    """

    # Check the image
    if len(input_img) != 2048:
        raise ValueError("The input image is not a valide CHEC camera image.")

    # Copy the input flat ctapipe image and add one element with the NaN value in the end

    input_img_ext = np.zeros(input_img.shape[0] + 1)
    input_img_ext[:-1] = input_img[:]
    input_img_ext[-1] = np.nan

    # Make the output image
    img_2d = input_img_ext[img_map]

    return img_2d


def array_2d_to_chec(img_2d):
    # Flatten image and remove NaN values
    img_1d = img_2d[np.isfinite(img_2d)]

    return img_1d



from ctapipe.utils import get_dataset_path
from ctapipe.io.hessioeventsource import HESSIOEventSource
from ctapipe.io.eventseeker import EventSeeker
import pytest


def test_eventseeker():
    dataset = get_dataset_path("gamma_test.simtel.gz")
    kwargs = dict(config=None, tool=None, input_url=dataset)
    with HESSIOEventSource(**kwargs) as reader:
        seeker = EventSeeker(reader=reader)
        event = seeker[1]
        assert event.r0.tels_with_data == {11, 21, 24, 26, 61, 63, 118, 119}
        event = seeker[0]
        assert event.r0.tels_with_data == {38, 47}
        event = seeker['409']
        assert event.r0.tels_with_data == {11, 21, 24, 26, 61, 63, 118, 119}
        tel_list = [{38, 47}, {11, 21, 24, 26, 61, 63, 118, 119}]
        events = seeker[0:2]
        events_tels = [e.r0.tels_with_data for e in events]
        assert events_tels == tel_list
        events = seeker[[0, 1]]
        events_tels = [e.r0.tels_with_data for e in events]
        assert events_tels == tel_list
        events = seeker[['408', '409']]
        events_tels = [e.r0.tels_with_data for e in events]
        assert events_tels == tel_list
        assert len(seeker) == 9

        with pytest.raises(IndexError):
            event = seeker[200]
            assert event is not None
        with pytest.raises(ValueError):
            event = seeker['t']
            assert event is not None
        with pytest.raises(TypeError):
            event = seeker[dict()]
            assert event is not None

    with HESSIOEventSource(**kwargs, max_events=5) as reader:
        seeker = EventSeeker(reader=reader)
        with pytest.raises(IndexError):
            event = seeker[5]
            assert event is not None

    class StreamFileReader(HESSIOEventSource):

        def is_stream(self):
            return True
    with StreamFileReader(**kwargs) as reader:
        with pytest.raises(IOError):
            seeker = EventSeeker(reader=reader)
            assert seeker is not None



"""
Image Cleaning Algorithms (identification of noisy pixels)
"""

__all__ = ['tailcuts_clean', 'dilate']

import numpy as np


def tailcuts_clean(geom, image, picture_thresh=7, boundary_thresh=5,
                   keep_isolated_pixels=False,
                   min_number_picture_neighbors=0):

    """Clean an image by selection pixels that pass a two-threshold
    tail-cuts procedure.  The picture and boundary thresholds are
    defined with respect to the pedestal dispersion. All pixels that
    have a signal higher than the picture threshold will be retained,
    along with all those above the boundary threshold that are
    neighbors of a picture pixel.

    To include extra neighbor rows of pixels beyond what are accepted, use the
    `ctapipe.image.dilate` function.

    Parameters
    ----------
    geom: `ctapipe.instrument.CameraGeometry`
        Camera geometry information
    image: array
        pixel values
    picture_thresh: float or array
        threshold above which all pixels are retained
    boundary_thresh: float or array
        threshold above which pixels are retained if they have a neighbor
        already above the picture_thresh
    keep_isolated_pixels: bool
        If True, pixels above the picture threshold will be included always,
        if not they are only included if a neighbor is in the picture or
        boundary
    min_number_picture_neighbors: int
        A picture pixel survives cleaning only if it has at least this number
        of picture neighbors. This has no effect in case keep_isolated_pixels is True

    Returns
    -------

    A boolean mask of *clean* pixels.  To get a zero-suppressed image and pixel
    list, use `image[mask], geom.pix_id[mask]`, or to keep the same
    image size and just set unclean pixels to 0 or similar, use
    `image[~mask] = 0`

    """
    pixels_above_picture = image >= picture_thresh

    if keep_isolated_pixels or min_number_picture_neighbors == 0:
        pixels_in_picture = pixels_above_picture
    else:
        # Require at least min_number_picture_neighbors. Otherwise, the pixel
        #  is not selected
        number_of_neighbors_above_picture = geom.neighbor_matrix_sparse.dot(
            pixels_above_picture.view(np.byte))
        pixels_in_picture = pixels_above_picture & (
            number_of_neighbors_above_picture >= min_number_picture_neighbors
        )

    # by broadcasting together pixels_in_picture (1d) with the neighbor
    # matrix (2d), we find all pixels that are above the boundary threshold
    # AND have any neighbor that is in the picture
    pixels_above_boundary = image >= boundary_thresh
    pixels_with_picture_neighbors = geom.neighbor_matrix_sparse.dot(
        pixels_in_picture)
    if keep_isolated_pixels:
        return (pixels_above_boundary
                & pixels_with_picture_neighbors) | pixels_in_picture
    else:
        pixels_with_boundary_neighbors = geom.neighbor_matrix_sparse.dot(
            pixels_above_boundary)
        return ((pixels_above_boundary & pixels_with_picture_neighbors) |
                (pixels_in_picture & pixels_with_boundary_neighbors))


def dilate(geom, mask):
    """
    Add one row of neighbors to the True values of a pixel mask and return
    the new mask.
    This can be used to include extra rows of pixels in a mask that was
    pre-computed, e.g. via `tailcuts_clean`.

    Parameters
    ----------
    geom: `~ctapipe.instrument.CameraGeometry`
        Camera geometry information
    mask: ndarray
        input mask (array of booleans) to be dilated
    """
    return mask | geom.neighbor_matrix_sparse.dot(mask)



__author__ = "@tino-michael"

import logging

import numpy as np
from astropy import units as u
from numba import jit

from ctapipe.instrument import CameraGeometry

logger = logging.getLogger(__name__)

__all__ = [
    "convert_geometry_hex1d_to_rect2d",
    "convert_geometry_rect2d_back_to_hexe1d"
]


def unskew_hex_pixel_grid(pix_x, pix_y, cam_angle=0 * u.deg,
                          base_angle=60 * u.deg):
    r"""transform the pixel coordinates of a hexagonal image into an
    orthogonal image

    Parameters
    ----------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the hexagonal pixel grid
    cam_angle : astropy.Quantity (default: 0 degrees)
        The skewing is performed along the y-axis, therefore, one of the slanted
        base-vectors needs to be parallel to the y-axis.
        Some camera grids are rotated in a way that this is not the case.
        This needs to be corrected.
    base_angle : astropy.Quantity (default: 60 degrees)
        the skewing angle of the hex-grid. should be 60° for regular hexagons

    Returns
    -------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the slanted, orthogonal pixel grid

    Notes
    -----
    The correction on the pixel position r can be described by a rotation R around
    one angle and a sheer S along a certain axis:

    .. math::
        r' = S \cdot R \cdot r

    .. math::
        \begin{pmatrix}
            x' \\
            y'
        \end{pmatrix}
        =
        \begin{pmatrix}
            1        &  0 \\
            -1/\tan  &  1
        \end{pmatrix}
        \cdot
        \begin{pmatrix}
            \cos  & -\sin \\
            \sin  &  \cos
        \end{pmatrix}
        \cdot
        \begin{pmatrix}
            x \\
            y
        \end{pmatrix}

    .. math::
        \begin{pmatrix}
            x' \\
            y'
        \end{pmatrix}
        =
        \begin{pmatrix}
                 \cos      &     -\sin      \\
            \sin-\cos/\tan & \sin/\tan+\cos
        \end{pmatrix}
        \cdot
        \begin{pmatrix}
            x \\
            y
        \end{pmatrix}

    """

    tan_angle = np.tan(base_angle)

    # If the camera-rotation angle is non-zero, create a rotation+sheering
    # matrix for the pixel coordinates
    if cam_angle != 0 * u.deg:
        sin_angle = np.sin(cam_angle)
        cos_angle = np.cos(cam_angle)

        # the correction on the pixel position r can be described by a
        # rotation R around one angle and a sheer S along a certain axis:
        #
        #  r'  = S * R * r
        # (x') = (   1    0) * (cos -sin) * (x) = (    cos         -sin    ) * (x)
        # (y')   (-1/tan  1)   (sin  cos)   (y)   (sin-cos/tan  sin/tan+cos) * (y)
        rot_mat = np.array(
            [[cos_angle, -sin_angle],
             [sin_angle - cos_angle / tan_angle,
              sin_angle / tan_angle + cos_angle]])

    else:
        # if we don't rotate the camera, only perform the sheer
        rot_mat = np.array([[1, 0], [-1 / tan_angle, 1]])

    rotated = np.dot(rot_mat, [pix_x.value, pix_y.value])
    rot_x = rotated[0] * pix_x.unit
    rot_y = rotated[1] * pix_x.unit
    return rot_x, rot_y


def reskew_hex_pixel_grid(pix_x, pix_y, cam_angle=0 * u.deg,
                          base_angle=60 * u.deg):
    r"""skews the orthogonal coordinates back to the hexagonal ones

    Parameters
    ----------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the slanted, orthogonal pixel grid
    cam_angle : astropy.Quantity (default: 0 degrees)
        The skewing is performed along the y-axis, therefore, one of the slanted
        base-vectors needs to be parallel to the y-axis.
        Some camera grids are rotated in a way that this is not the case.
        This needs to be corrected.
    base_angle : astropy.Quantity (default: 60 degrees)
        the skewing angle of the hex-grid. should be 60° for regular hexagons

    Returns
    -------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the hexagonal pixel grid

    Notes
    -----
    To revert the rotation, we need to find matrices S' and R' with
    :math:`S' \cdot S = 1` and :math:`R' \cdot R = 1`,
    so that :math:`r = R' \cdot S' \cdot S \cdot R \cdot r = R' \cdot S' \cdot  r'`:

    .. math::
        \begin{pmatrix}
            x \\
            y
        \end{pmatrix}
        =
        \begin{pmatrix}
            \cos  &  \sin \\
            -\sin &  \cos
        \end{pmatrix}
        \cdot
        \begin{pmatrix}
            1       &  0 \\
            1/\tan  &  1
        \end{pmatrix}
        \cdot
        \begin{pmatrix}
            x' \\
            y'
        \end{pmatrix}

    .. math::
        \begin{pmatrix}
            x \\
            y
        \end{pmatrix}
        =
        \begin{pmatrix}
            \cos+\sin/\tan  &  \sin \\
            \cos/\tan-\sin  &  \cos
        \end{pmatrix}
        \cdot
        \begin{pmatrix}
            x' \\
            y'
        \end{pmatrix}

    """

    tan_angle = np.tan(base_angle)

    # If the camera-rotation angle is non-zero, create a rotation+sheering
    # matrix for the pixel coordinates
    if cam_angle != 0 * u.deg:
        sin_angle = np.sin(cam_angle)
        cos_angle = np.cos(cam_angle)

        # to revert the rotation, we need to find matrices S' and R'
        # S' * S = 1 and R' * R = 1
        # so that
        # r = R' * S' * S * R * r = R' * S' *  r'
        #
        # (x) = ( cos sin) * (  1    0) * (x') = (cos+sin/tan  sin) * (x')
        # (y)   (-sin cos)   (1/tan  1)   (y')   (cos/tan-sin  cos)   (y')

        rot_mat = np.array(
            [[cos_angle + sin_angle / tan_angle, sin_angle],
             [cos_angle / tan_angle - sin_angle, cos_angle]])

    else:
        # if we don't rotate the camera, only perform the sheer
        rot_mat = np.array([[1, 0], [1 / tan_angle, 1]])

    rotated = np.dot(rot_mat, [pix_x.value, pix_y.value])
    rot_x = rotated[0] * pix_x.unit
    rot_y = rotated[1] * pix_x.unit
    return rot_x, rot_y


@jit
def reskew_hex_pixel_from_orthogonal_edges(x_edges, y_edges, square_mask):
    """extracts and skews the pixel coordinates from a 2D orthogonal
    histogram (i.e. the bin-edges) and skews them into the hexagonal
    image while selecting only the pixel that are selected by the
    given mask

    Parameters
    ----------
    x_edges, y_edges : 1darrays
        the bin edges of the 2D histogram
    square_mask : 2darray
        mask that selects the pixels actually belonging to the camera

    Returns
    -------
    unrot_x, unrot_y : 1darrays
        pixel coordinated reskewed into the hexagonal camera grid
    """

    unrot_x, unrot_y = [], []
    for i, x in enumerate((x_edges[:-1] + x_edges[1:]) / 2):
        for j, y in enumerate((y_edges[:-1] + y_edges[1:]) / 2):
            if square_mask[i][j]:
                x_unrot, y_unrot = reskew_hex_pixel_grid(x, y)
                unrot_x.append(x_unrot)
                unrot_y.append(y_unrot)
    return unrot_x, unrot_y


@jit
def get_orthogonal_grid_edges(pix_x, pix_y, scale_aspect=True):
    """calculate the bin edges of the slanted, orthogonal pixel grid to
    resample the pixel signals with np.histogramdd right after.

    Parameters
    ----------
    pix_x, pix_y : 1D numpy arrays
        the list of x and y coordinates of the slanted, orthogonal pixel grid
    scale_aspect : boolean (default: True)
        if True, rescales the x-coordinates to create square pixels
        (instead of rectangular ones)

    Returns
    --------
    x_edges, y_edges : 1D numpy arrays
        the bin edges for the slanted, orthogonal pixel grid
    x_scale : float
        factor by which the x-coordinates have been scaled
    """

    # finding the size of the square patches
    d_x = 99 * u.m
    d_y = 99 * u.m
    x_base = pix_x[0]
    y_base = pix_y[0]
    for x, y in zip(pix_x, pix_y):
        if abs(y - y_base) < abs(x - x_base):
            d_x = min(d_x, abs(x - x_base))
    for x, y in zip(pix_x, pix_y):
        if abs(y - y_base) > abs(x - x_base):
            d_y = min(d_y, abs(y - y_base))

    x_scale = 1
    if scale_aspect:
        x_scale = d_y / d_x
        pix_x *= x_scale
        d_x = d_y

    # with the maximal extension of the axes and the size of the pixels,
    # determine the number of bins in each direction
    n_bins_x = np.around(abs(max(pix_x) - min(pix_x)) / d_x) + 2
    n_bins_y = np.around(abs(max(pix_y) - min(pix_y)) / d_y) + 2
    x_edges = np.linspace(min(pix_x).value, max(pix_x).value, n_bins_x)
    y_edges = np.linspace(min(pix_y).value, max(pix_y).value, n_bins_y)

    return x_edges, y_edges, x_scale


rot_buffer = {}


def convert_geometry_hex1d_to_rect2d(geom, signal, key=None, add_rot=0):
    """converts the geometry object of a camera with a hexagonal grid into
    a square grid by slanting and stretching the 1D arrays of pixel x
    and y positions and signal intensities are converted to 2D
    arrays. If the signal array contains a time-dimension it is
    conserved.

    Parameters
    ----------
    geom : CameraGeometry object
        geometry object of hexagonal cameras
    signal : ndarray
        1D (no timing) or 2D (with timing) array of the pmt signals
    key : (default: None)
        arbitrary key to store the transformed geometry in a buffer
    add_rot : int/float (default: 0)
        parameter to apply an additional rotation of `add_rot` times 60°

    Returns
    -------
    new_geom : CameraGeometry object
        geometry object of the slanted picture now with a rectangular
        grid and a 2D grid for the pixel positions. contains now a 2D
        masking array signifying which of the pixels came from the
        original geometry and which are simply fillers from the
        rectangular grid
    rot_img : ndarray 2D (no timing) or 3D (with timing)
        the rectangular signal image
    """

    if key in rot_buffer:

        # if the conversion with this key was done before and stored,
        # just read it in
        (geom, new_geom, hex_to_rect_map) = rot_buffer[key]
    else:

        # otherwise, we have to do the conversion first now,
        # skew all the coordinates of the original geometry

        # extra_rot is the angle to get back to aligned hexagons with flat
        # tops. Note that the pixel rotation angle brings the camera so that
        # hexagons have a point at the top, so need to go 30deg back to
        # make them flat
        extra_rot = geom.pix_rotation - 30 * u.deg

        # total rotation angle:
        rot_angle = (add_rot * 60 * u.deg) - extra_rot

        logger.debug("geom={}".format(geom))
        logger.debug("rot={}, extra={}".format(rot_angle, extra_rot))

        rot_x, rot_y = unskew_hex_pixel_grid(geom.pix_x, geom.pix_y,
                                             cam_angle=rot_angle)

        # with all the coordinate points, we can define the bin edges
        # of a 2D histogram
        x_edges, y_edges, x_scale = get_orthogonal_grid_edges(rot_x, rot_y)

        # this histogram will introduce bins that do not correspond to
        # any pixel from the original geometry. so we create a mask to
        # remember the true camera pixels by simply throwing all pixel
        # positions into numpy.histogramdd: proper pixels contain the
        # value 1, false pixels the value 0.
        square_mask = np.histogramdd([rot_y, rot_x],
                                     bins=(y_edges, x_edges))[0].astype(bool)

        # to be consistent with the pixel intensity, instead of saving
        # only the rotated positions of the true pixels (rot_x and
        # rot_y), create 2D arrays of all x and y positions (also the
        # false ones).
        grid_x, grid_y = np.meshgrid((x_edges[:-1] + x_edges[1:]) / 2.,
                                     (y_edges[:-1] + y_edges[1:]) / 2.)

        ids = []
        # instead of blindly enumerating all pixels, let's instead
        # store a list of all valid -- i.e. picked by the mask -- 2D
        # indices
        for i, row in enumerate(square_mask):
            for j, val in enumerate(row):
                if val is True:
                    ids.append((i, j))

        # the area of the pixels (note that this is still a deformed
        # image)
        pix_area = (np.ones_like(grid_x)
                    * (x_edges[1] - x_edges[0])
                    * (y_edges[1] - y_edges[0]))

        # creating a new geometry object with the attributes we just determined
        new_geom = CameraGeometry(
            cam_id=geom.cam_id + "_rect",
            pix_id=ids,  # this is a list of all the valid coordinate pairs now
            pix_x=grid_x * u.m,
            pix_y=grid_y * u.m,
            pix_area=pix_area * u.m ** 2,
            neighbors=geom.neighbors,
            pix_type='rectangular', apply_derotation=False)

        # storing the pixel mask for later use
        new_geom.mask = square_mask

        # create a transfer map by enumerating all pixel positions in a 2D histogram
        hex_to_rect_map = np.histogramdd([rot_y, rot_x],
                                         bins=(y_edges, x_edges),
                                         weights=np.arange(len(signal)))[
            0].astype(int)
        # bins that do not correspond to the original image get an entry of `-1`
        hex_to_rect_map[~square_mask] = -1

        if signal.ndim > 1:
            long_map = []
            for i in range(signal.shape[-1]):
                tmp_map = hex_to_rect_map + i * (np.max(hex_to_rect_map) + 1)
                tmp_map[~square_mask] = -1
                long_map.append(tmp_map)
            hex_to_rect_map = np.array(long_map)

        if key is not None:
            # if a key is given, store the essential objects in a buffer
            rot_buffer[key] = (geom, new_geom, hex_to_rect_map)

    # done `if key in rot_buffer`

    # create the rotated rectangular image by applying `hex_to_rect_map` to the flat,
    # extended input image
    # `input_img_ext` is the flattened input image extended by one entry that contains NaN
    # since `hex_to_rect_map` contains `-1` for "fake" pixels, it maps this extra NaN
    # value at the last array position to any bin that does not correspond to a pixel of
    # the original image
    input_img_ext = np.full(np.prod(signal.shape) + 1, np.nan)

    # the way the map is produced, it has the time dimension as axis=0;
    # but `signal` has it as axis=-1, so we need to roll the axes back and forth a bit.
    # if there is no time dimension, `signal` is a 1d array and `rollaxis` has no effect.
    input_img_ext[:-1] = np.rollaxis(signal, axis=-1, start=0).ravel()

    # now apply the transfer map
    rot_img = input_img_ext[hex_to_rect_map]

    # if there is a time dimension, roll the time axis back to the last position
    try:
        rot_img = np.rollaxis(rot_img, 0, 3)
    except ValueError:
        pass

    return new_geom, rot_img


def convert_geometry_rect2d_back_to_hexe1d(geom, signal, key=None,
                                           add_rot=None):
    """reverts the geometry distortion performed by convert_geometry_hexe1d_to_rect_2d
    back to a hexagonal grid stored in 1D arrays

    Parameters
    ----------
    geom : CameraGeometry
        geometry object where pixel positions are stored in a 2D
        rectangular camera grid
    signal : ndarray
        pixel intensity stored in a 2D rectangular camera grid
    key:
        key to retrieve buffered geometry information
    add_rot:
        not used -- only here for backwards compatibility

    Returns
    -------
    old_geom : CameraGeometry
        the original geometry of the image
    signal : ndarray
        1D (no timing) or 2D (with timing) array of the pmt signals

    Notes
    -----
    The back-conversion works with an internal buffer to store the transfer map (which
    was produced in the first conversion). If `key` is not found in said buffer, this
    function tries to perform a mock conversion. For this, it needs a `CameraGeometry`
    instance of the original camera layout, which it tries to load by name (i.e.
    the `cam_id`). The function assumes the original `cam_id` can be inferred from the
    given, modified one by: `geom.cam_id.split('_')[0]`.
    """

    if key not in rot_buffer:
        # if the key is not in the buffer from the initial conversion (maybe
        # because you did it in another process?), perform a mock conversion
        # here ATTENTION assumes the original cam_id can be inferred from the
        #  given, modified one by by `geom.cam_id.split('_')[0]`
        try:
            orig_geom = CameraGeometry.from_name(geom.cam_id.split('_')[0])
        except:
            raise ValueError(
                "could not deduce `CameraGeometry` from given `geom`...\n"
                "please provide a `geom`, so that "
                "`geom.cam_id.split('_')[0]` is a known `cam_id`")

        orig_signal = np.zeros(len(orig_geom.pix_x))
        convert_geometry_hex1d_to_rect2d(geom=orig_geom, signal=orig_signal,
                                         key=key, add_rot=add_rot)

    (old_geom, new_geom, hex_square_map) = rot_buffer[key]

    # the output image has as many entries as there are non-negative values in the
    # transfer map (this accounts for time as well)
    unrot_img = np.zeros(np.count_nonzero(hex_square_map >= 0))

    # rearrange input `signal` according to the mask and map
    # (the dots in the brackets expand the mask to account for a possible time dimension)
    # `atleast_3d` ensures that there is a third axis that we can roll to the front
    # even if there is no time; if we'd use `axis=-1` instead, in cas of no time
    # dimensions, we would rotate the x and y axes, resulting in a mirrored image
    # `squeeze` reduces the added axis again in the no-time-slices cases
    unrot_img[hex_square_map[..., new_geom.mask]] = \
        np.squeeze(np.rollaxis(np.atleast_3d(signal), 2, 0))[..., new_geom.mask]

    # if `signal` has a third dimension, that is the time
    # and we need to roll some axes again...
    if signal.ndim == 3:
        # unrot_img[hex_square_map[..., new_geom.mask]] = \
        # np.rollaxis(signal, -1, 0)[..., new_geom.mask]

        # reshape the image so that the time is the first axis
        # and then roll the time to the back
        unrot_img = unrot_img.reshape((signal.shape[2],
                                       np.count_nonzero(new_geom.mask)))
        unrot_img = np.rollaxis(unrot_img, -1, 0)
    # else:
    #     unrot_img[hex_square_map[new_geom.mask]] = \
    #         signal[new_geom.mask]

    return old_geom, unrot_img



from .hillas import *
from .cleaning import *
from .pixel_likelihood import *
from .charge_extractors import *
from .waveform_cleaning import *
from .reducers import *
from .muon import *
from .geometry_converter import *



from ctapipe.core import Component
from time import sleep


class Inverse(Component):
    """Add class represents a Stage for pipeline.
    It returns inverted value of received value
    """

    def init(self):
        self.log.debug("--- Add init ---")
        return True

    def run(self, inputval):
        sleep(.5)
        if input:
            self.log.debug("Inverse received %s", input)
            return int(inputval) * -1

    def finish(self):
        self.log.debug("--- Add finish ---")



from copy import copy, deepcopy
from ctapipe.core.component import Component
from inspect import isabstract
from traitlets.config.loader import Config
from traitlets import CaselessStrEnum


class FactoryMeta(type(Component), type):
    def __new__(mcs, name, bases, dct):

        # Setup class lookup
        base = dct['base']
        dct['subclasses'] = None
        dct['subclass_names'] = None
        if base:
            if not isinstance(base, type) or not issubclass(base, Component):
                raise AttributeError("Factory.base must be set to a Component")

            dct['subclasses'] = mcs.child_subclasses(base)
            dct['subclass_names'] = [c.__name__ for c in dct['subclasses']]

            default = None if 'default' not in dct else dct['default']
            help_msg = 'Product class to obtain from the Factory.'
            if 'custom_product_help' in dct and dct['custom_product_help']:
                help_msg = dct['custom_product_help']
            dct['product'] = CaselessStrEnum(
                dct['subclass_names'],
                default,
                allow_none=True,
                help=help_msg
            ).tag(config=True)

            # Gather a record of which traits are valid for which subclasses
            # and copy subclass traits
            traits = dict()
            record = dict()
            for sub in dct['subclasses']:
                for key, trait in sub.class_traits().items():
                    if key in ['config', 'parent']:
                        continue
                    record.setdefault(key, []).append(sub.__name__)
                    if key not in traits:
                        traits[key] = deepcopy(trait)

            # Add subclass traits to Factory and include a list of valid
            # sublasses for the trait help message
            for key, trait in traits.items():
                trait.help += "\n\nCompatible Components: " + str(record[key])
                dct[key] = trait

        return type.__new__(mcs, name, bases, dct)

    @staticmethod
    def child_subclasses(base):
        """
        Return all non-abstract subclasses of a base class.

        Parameters
        ----------
        base : class
            high level class object that is inherited by the
            desired subclasses

        Returns
        -------
        children : list
            list of non-abstract subclasses

        """
        family = base.__subclasses__() + [
            g for s in base.__subclasses__()
            for g in FactoryMeta.child_subclasses(s)
        ]
        children = [g for g in family if not isabstract(g)]

        return children


class Factory(Component, metaclass=FactoryMeta):
    """
    A base class for all class factories that exist in the `Tools`/`Components`
    frameworks.

    To create a Factory, inherit this class and set `base` to the
    base-class of the Factory.

    The traits of the sub-classes are automatically added to the Factory
    (and are included in the help message).

    The traits and kwargsthat correctly correspond to the product are passed
    onto the product.

    To use a factory from within a `ctapipe.core.tool.Tool`:

    >>> cls = Factory.produce(config=self.config, tool=self)

    Attributes
    ----------
    base : type
        The base-class for the different classes that could be returned from
        the factory
    product : traitlets.CaselessStrEnum
        The traitlet allowing the manual specification of which class is
        returned from the factory. This is created in `FactoryMeta`.
    subclasses : list
        A list of the subclasses for the base-class specified in `base`
    subclass_names : list
        The names of the classes in `subclasses`
    default : str
        The default product to return. If None, then there is no default.
    custom_product_help : None or str
        If set, then this text will be displayed for the help message for the
        product traitlet.
    """
    base = None
    product = None  # Instanced as a traitlet by FactoryMeta
    subclasses = None  # Filled by FactoryMeta
    subclass_names = None  # Filled by FactoryMeta
    default = None
    custom_product_help = None

    def __init__(self, config=None, tool=None, **kwargs):
        """
        Base Factory class

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs

        """
        if not self.base:
            raise AttributeError("Factory.base must be set to a Component")

        super().__init__(config=config, parent=tool, **kwargs)
        self.kwargs = copy(kwargs)

    def _get_product_name(self):
        """
        Method to obtain the correct name for the product.

        Returns
        -------
        str
            The name of the product to return from the Factory.
        """
        if self.product:
            return self.product
        else:
            raise AttributeError("The user has not specified a product for {}"
                                 .format(self.__class__.__name__))

    @property
    def _product(self):
        """
        Obtain the class constructor for the specified product name.

        Returns
        -------
        product : class

        """
        subclass_dict = dict(zip(self.subclass_names, self.subclasses))
        product_name = self._get_product_name()
        self.log.info("Obtaining {} from {}".format(product_name,
                                                    self.__class__.__name__))
        try:
            product = subclass_dict[product_name]
            return product
        except KeyError:
            self.log.exception('No product found with name "{}" for '
                               'factory.'.format(product_name))
            raise

    @property
    def _instance(self):
        """
        Obtain an instance of the product class with the config and arguments
        correctly copied from the Factory to the product.

        Returns
        -------
        instance
            Instance of the product class that is the purpose of the factory
            to produce.
        """
        product = self._product
        product_traits = product.class_trait_names()
        product_args = list(product.__init__.__code__.co_varnames)
        config = copy(self.__dict__['_trait_values']['config'])
        parent = copy(self.__dict__['_trait_values']['parent'])

        # Update traitlet defaults (they may have been changed via the Factory)
        for name, trait in self.class_own_traits().items():
            for sub in self.subclasses:
                try:
                    sub_trait = sub.class_traits()[name]
                    sub_trait.default_value = trait.default_value
                except KeyError:
                    pass

        if config[self.__class__.__name__]:
            # If Product config does not exist, create new Config instance
            # Note: `config[product.__name__]` requires Config, not dict
            if not config[product.__name__]:
                config[product.__name__] = Config()

            # Copy Factory config to Product config
            for key, value in config[self.__class__.__name__].items():
                if key in product_traits:
                    config[product.__name__][key] = value

        # Copy valid arguments to kwargs
        kwargs = copy(self.kwargs)
        for key in list(kwargs.keys()):
            if key not in product_traits + product_args:
                del kwargs[key]

        instance = product(config, parent, **kwargs)
        return instance

    @classmethod
    def produce(cls, config=None, tool=None, **kwargs):
        """
        Produce an instance of the product class

        Parameters
        ----------
        config : traitlets.loader.Config
            Configuration specified by config file or cmdline arguments.
            Used to set traitlet values.
            Set to None if no configuration to pass.
        tool : ctapipe.core.Tool
            Tool executable that is calling this component.
            Passes the correct logger to the component.
            Set to None if no Tool to pass.
        kwargs

        Returns
        -------
        instance
            Instance of the product class that is the purpose of the factory
            to produce.

        """
        factory = cls(config=config, tool=tool, **kwargs)
        instance = factory._instance
        return instance



from traitlets import (Int, Integer, Float, Unicode, Enum, Long, List,
                       Bool, CRegExp, Dict, TraitError, observe,
                       CaselessStrEnum, TraitType)
from traitlets.config import boolean_flag as flag
import os

__all__ = ['Path', 'Int', 'Integer', 'Float', 'Unicode', 'Enum', 'Long', 'List',
           'Bool', 'CRegExp', 'Dict', 'flag', 'TraitError', 'observe',
           'CaselessStrEnum']


class Path(TraitType):
    def __init__(self, exists=None, directory_ok=True, file_ok=True):
        '''
        A path Trait for input/output files.

        Parameters
        ----------
        exists: boolean or None
            If True, path must exist, if False path must not exist

        directory_ok: boolean
            If False, path must not be a directory
        file_ok: boolean
            If False, path must not be a file
        '''
        super().__init__()
        self.exists = exists
        self.directory_ok = directory_ok
        self.file_ok = file_ok

    def validate(self, obj, value):

        if isinstance(value, str):
            value = os.path.abspath(value)
            if self.exists is not None:
                if os.path.exists(value) != self.exists:
                    raise TraitError('Path "{}" {} exist'.format(
                        value,
                        'does not' if self.exists else 'must'
                    ))
            if os.path.exists(value):
                if os.path.isdir(value) and not self.directory_ok:
                    raise TraitError(
                        'Path "{}" must not be a directory'.format(value)
                    )
                if os.path.isfile(value) and not self.file_ok:
                    raise TraitError(
                        'Path "{}" must not be a file'.format(value)
                    )

            return value

        return self.error(obj, value)



"""
Algorithms to select correct gain channel
"""
from abc import ABCMeta, abstractclassmethod

import numpy as np

from ...core import Component, Factory, traits
from ...utils import get_table_dataset

__all__ = ['GainSelectorFactory',
           'ThresholdGainSelector',
           'SimpleGainSelector',
           'pick_gain_channel']


def pick_gain_channel(waveforms, threshold, select_by_sample=False):
    """
    the PMTs on some cameras have 2 gain channels. select one
    according to a threshold.

    Parameters:
    -----------
    waveforms: np.ndarray
        Array of shape (N_gain, N_pix, N_samp)
    threshold: float
        threshold (in PE/sample) of when to switch to low-gain channel
    select_by_sample: bool
        if true, select only low-gain *samples* when high-gain is over
        threshold

    Returns
    -------
    tuple:
        gain-selected intensity, boolean array of which channel was chosen
    """

    # if we have 2 channels:
    if waveforms.shape[0] == 2:
        waveforms = np.squeeze(waveforms)
        new_waveforms = waveforms[0].copy()

        if select_by_sample:
            # replace any samples that are above threshold with low-gain ones:
            gain_mask = waveforms[0] > threshold
            new_waveforms[gain_mask] = waveforms[1][gain_mask]
        else:
            # use entire low-gain waveform if any sample of high-gain
            # waveform is above threshold
            gain_mask = (waveforms[0] > threshold).any(axis=1)
            new_waveforms[gain_mask] = waveforms[1][gain_mask]

    elif waveforms.shape[0] == 1:
        new_waveforms = np.squeeze(waveforms)
        gain_mask = np.zeros_like(new_waveforms).astype(bool)

    else:
        raise ValueError("input waveforms has shape %s. not sure what to do "
                         "with that.", waveforms.shape)

    return new_waveforms, gain_mask


class GainSelector(Component, metaclass=ABCMeta):
    """
    Base class for algorithms that reduce a 2-gain-channel waveform to a
    single waveform.
    """

    @abstractclassmethod
    def select_gains(self, cam_id, multi_gain_waveform):
        """
        Takes an input waveform and cam_id  and performs gain selection

        Returns
        -------
        tuple(ndarray, ndarray):
            (waveform, gain_mask), where the gain_mask is a boolean array of
            which gain channel was used.
        """
        pass


class NullGainSelector(GainSelector):
    """
    do no gain selection, leaving possibly 2 gain channels at the DL1 level.
    this may break further steps in the chain if they do not expect 2 gains.
    """

    def select_gains(self, cam_id, multi_gain_waveform):
        return multi_gain_waveform, np.ones(multi_gain_waveform.shape[1])


class SimpleGainSelector(GainSelector):
    """
    Simply choose a single gain channel always.
    """

    channel = traits.Int(default_value=0, help="which gain channel to "
                                               "retain").tag(config=True)

    def select_gains(self, cam_id, multi_gain_waveform):
        return (
            multi_gain_waveform[self.channel],
            (np.ones(multi_gain_waveform.shape[1]) * self.channel).astype(
                np.bool)
        )


class ThresholdGainSelector(GainSelector):
    """
    Select gain channel using fixed-threshold for any sample in the waveform.
    The thresholds are loaded from an `astropy.table.Table` that must contain
    two columns: `cam_id` (the name of the camera) and `gain_threshold_pe`,
    the threshold in photo-electrons per sample at which the switch should
    occur.

    Parameters
    ----------
    threshold_table_name: str
        Name of gain channel table to load
    select_by_sample: bool
        If True, replaces only the waveform samples that are above
        the threshold with low-gain versions, otherwise the full
        low-gain waveform is used.

    Attributes
    ----------
    thresholds: dict
        mapping of cam_id to threshold value
    """

    threshold_table_name = traits.Unicode(
        default_value='gain_channel_thresholds',
        help='Name of gain channel table to load'
    ).tag(config=True)

    select_by_sample = traits.Bool(
        default_value=False,
        help='If True, replaces only the waveform samples that are above '
             'the threshold with low-gain versions, otherwise the full '
             'low-gain waveform is used.'
    ).tag(config=True)

    def __init__(self, config=None, parent=None, **kwargs):
        super().__init__(config=config, parent=parent, **kwargs)

        tab = get_table_dataset(
            self.threshold_table_name,
            role='dl0.tel.svc.gain_thresholds'
        )
        self.thresholds = dict(zip(tab['cam_id'], tab['gain_threshold_pe']))
        self.log.debug("Loaded threshold table: \n %s", tab)

    def __str__(self):
        return "{}({})".format(self.__class__.__name__, self.thresholds)

    def select_gains(self, cam_id, multi_gain_waveform):

        try:
            threshold = self.thresholds[cam_id]
        except KeyError:
            raise KeyError(
                "Camera ID '{}' not found in the gain-threshold "
                "table '{}'".format(cam_id, self.threshold_table_name)
            )

        waveform, gain_mask = pick_gain_channel(
            waveforms=multi_gain_waveform,
            threshold=threshold,
            select_by_sample=self.select_by_sample
        )

        return waveform, gain_mask


class GainSelectorFactory(Factory):
    """
    Factory to obtain a GainSelector
    """
    base = GainSelector
    default = 'ThresholdGainSelector'
    custom_product_help = 'Gain-channel selection scheme to use.'



from tempfile import TemporaryDirectory

import numpy as np
from sklearn.neural_network import MLPClassifier
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from ctapipe.reco.event_classifier import EventClassifier


def test_pipeline_classifier():
    cam_id_list = ["FlashCam", "ASTRICam"]
    feature_list = {"FlashCam": [[1, 10], [2, 20], [3, 30], [0.9, 9],
                                 [10, 1], [20, 2], [30, 3], [9, 0.9]],
                    "ASTRICam": [[10, 1], [20, 2], [30, 3], [9, 0.9],
                                 [1, 10], [2, 20], [3, 30], [0.9, 9]]}
    target_list = {"FlashCam": ["a", "a", "a", "a", "b", "b", "b", "b"],
                   "ASTRICam": ["a", "a", "a", "a", "b", "b", "b", "b"]}

    estimators = [('scaler', StandardScaler()),
                  ('clf', MLPClassifier(max_iter=400))]

    clf = EventClassifier(classifier=Pipeline, steps=estimators,
                          cam_id_list=cam_id_list)
    clf.fit(feature_list, target_list)

    prediction = clf.predict_by_event([{"ASTRICam": [[10, 1]]},
                                       {"ASTRICam": [[2, 20]]},
                                       {"ASTRICam": [[3, 30]]}])
    assert (prediction == ["a", "b", "b"]).all()

    prediction = clf.predict_by_event([{"FlashCam": [[10, 1]]},
                                       {"FlashCam": [[2, 20]]},
                                       {"FlashCam": [[3, 30]]}])
    assert (prediction == ["b", "a", "a"]).all()


def test_prepare_model_MLP():
    cam_id_list = ["FlashCam", "ASTRICam"]
    feature_list = {"FlashCam": [[1, 10], [2, 20], [3, 30], [0.9, 9],
                                 [10, 1], [20, 2], [30, 3], [9, 0.9]],
                    "ASTRICam": [[10, 1], [20, 2], [30, 3], [9, 0.9],
                                 [1, 10], [2, 20], [3, 30], [0.9, 9]]}
    target_list = {"FlashCam": ["a", "a", "a", "a", "b", "b", "b", "b"],
                   "ASTRICam": ["a", "a", "a", "a", "b", "b", "b", "b"]}

    clf = EventClassifier(classifier=MLPClassifier, cam_id_list=cam_id_list, max_iter=400)
    scaled_features, scaler = EventClassifier.scale_features(cam_id_list, feature_list)

    # clf.fit(feature_list, target_list)
    clf.fit(scaled_features, target_list)
    return clf, cam_id_list, scaler


def test_fit_save_load_MLP():
    clf, cam_id_list, scaler = test_prepare_model_MLP()
    with TemporaryDirectory() as d:
        temp_path = "/".join([d, "reg_{cam_id}.pkl"])
        clf.save(temp_path)
        clf = EventClassifier.load(temp_path, cam_id_list)
        return clf, cam_id_list, scaler


def test_predict_by_event_MLP():
    clf, cam_id_list, scaler = test_fit_save_load_MLP()

    x = scaler['ASTRICam'].transform(np.array([10, 1], dtype=float).reshape(1, -1))
    x = {'ASTRICam': x}
    y = scaler['ASTRICam'].transform(np.array([2, 20], dtype=float).reshape(1, -1))
    y = {'ASTRICam': y}
    z = scaler['ASTRICam'].transform(np.array([3, 30], dtype=float).reshape(1, -1))
    z = {'ASTRICam': z}
    prediction = clf.predict_by_event([x, y, z])
    assert (prediction == ["a", "b", "b"]).all()

    x = scaler['FlashCam'].transform(np.array([10, 1], dtype=float).reshape(1, -1))
    x = {'FlashCam': x}
    y = scaler['FlashCam'].transform(np.array([2, 20], dtype=float).reshape(1, -1))
    y = {'FlashCam': y}
    z = scaler['FlashCam'].transform(np.array([3, 30], dtype=float).reshape(1, -1))
    z = {'FlashCam': z}
    prediction = clf.predict_by_event([x, y, z])
    assert (prediction == ["b", "a", "a"]).all()


def test_prepare_model():
    cam_id_list = ["FlashCam", "ASTRICam"]
    feature_list = {"FlashCam": [[1, 10], [2, 20], [3, 30], [0.9, 9],
                                 [10, 1], [20, 2], [30, 3], [9, 0.9]],
                    "ASTRICam": [[10, 1], [20, 2], [30, 3], [9, 0.9],
                                 [1, 10], [2, 20], [3, 30], [0.9, 9]]}
    target_list = {"FlashCam": ["a", "a", "a", "a", "b", "b", "b", "b"],
                   "ASTRICam": ["a", "a", "a", "a", "b", "b", "b", "b"]}

    clf = EventClassifier(cam_id_list=cam_id_list)
    clf.fit(feature_list, target_list)
    return clf, cam_id_list


def test_fit_save_load():
    clf, cam_id_list = test_prepare_model()
    with TemporaryDirectory() as d:
        temp_path = "/".join([d, "reg_{cam_id}.pkl"])
        clf.save(temp_path)
        clf = EventClassifier.load(temp_path, cam_id_list)
        return clf, cam_id_list


def test_predict_by_event():
    clf, cam_id_list = test_fit_save_load()
    prediction = clf.predict_by_event([{"ASTRICam": [[10, 1]]},
                                       {"ASTRICam": [[2, 20]]},
                                       {"ASTRICam": [[3, 30]]}])
    assert (prediction == ["a", "b", "b"]).all()

    prediction = clf.predict_by_event([{"FlashCam": [[10, 1]]},
                                       {"FlashCam": [[2, 20]]},
                                       {"FlashCam": [[3, 30]]}])
    assert (prediction == ["b", "a", "a"]).all()


def test_Qfactor():
    """
    TODO: how to test validity of Q-factor values?
    """
    cam_id = ["ASTRICam"]
    features = {"ASTRICam": [[10, 1], [20, 2], [30, 3], [9, 0.9],
                             [1, 10], [2, 20], [3, 30], [0.9, 9]]}
    target = {"ASTRICam": [1, 1, 1, 1, 0, 0, 0, 0]}
    clf = EventClassifier(cam_id_list=cam_id)
    clf.fit(features, target)

    # Now predict
    ev_feat = [{"ASTRICam": [[10, 1]]}, {"ASTRICam": [[2, 20]]}, {"ASTRICam": [[3, 30]]},
               {"ASTRICam": [[100, 10]]}, {"ASTRICam": [[4, 40]]}, {"ASTRICam": [[0.5, 5]]}]
    true_labels = np.array([1, 0, 0, 1, 0, 0], dtype=np.int8)
    prediction = clf.predict_proba_by_event(X=ev_feat)

    # prediction is a two columns array
    # first  column is the probability to belong to class "0" --> hadron
    # second column is the probability to belong to class "1" --> gamma
    # we are interested in the probability to be gamma
    proba_to_be_gamma = prediction[:, 1]

    Q, gammaness = clf.compute_Qfactor(
        proba=proba_to_be_gamma, labels=true_labels, nbins=2)

    assert Q.size != 0
    assert Q.size == gammaness.size


def test_hyperBinning():
    clf = EventClassifier(cam_id_list=None)
    x = np.array([[26, 70, 53],
                  [97, 20, 56],
                  [35, 38, 81],
                  [48, 60, 40],
                  [73, 68, 63],
                  [96, 86, 63],
                  [73, 67, 6],
                  [48, 66, 60],
                  [47, 82, 87],
                  [60, 52, 74]])

    dum_l = [{'maxf': max(x[:, 0]), 'minf': min(x[:, 0]), 'col': 0, 'nbins': 4},
             {'maxf': max(x[:, 1]), 'minf': min(x[:, 1]), 'col': 1, 'nbins': 2}]

    dum_g = clf._hyperBinning(x, dum_l)

    assert np.all(dum_g.size() == (1, 1, 3, 2, 1))


def test_level_populations():
    clf = EventClassifier(cam_id_list=None)
    g = np.array([[26, 70, 53],
                  [97, 20, 56],
                  [35, 38, 81],
                  [48, 60, 40],
                  [73, 68, 63],
                  [96, 86, 63],
                  [73, 67, 6],
                  [48, 66, 60],
                  [47, 82, 87],
                  [60, 52, 74]])

    h = np.array([[18, 31, 47],
                  [15, 81, 72],
                  [75, 93, 45],
                  [57, 50, 3],
                  [12, 80, 3],
                  [82, 49, 31],
                  [1, 21, 0],
                  [79, 12, 29],
                  [19, 52, 42],
                  [86, 49, 15]])

    dum_l = [{'maxf': 100, 'minf': 0, 'col': 0, 'nbins': 4},
             {'maxf': 100, 'minf': 0, 'col': 1, 'nbins': 2}]

    group_g = clf._hyperBinning(g, dum_l)
    group_h = clf._hyperBinning(h, dum_l)

    cleaned_g, cleaned_h = clf.level_populations(group_g, group_h, g, h)

    assert cleaned_g.shape == cleaned_h.shape


